import unittest
from unittest import mock

import quilt3
from quilt3.telemetry import ApiTelemetry


@ApiTelemetry(mock.sentinel.API_NAME1)
def test_api_function1():
    pass


@ApiTelemetry(mock.sentinel.API_NAME2)
def test_api_function2():
    pass


class TelemetryTest(unittest.TestCase):
    def setUp(self):
        super().setUp()

        patcher = mock.patch('quilt3.telemetry.ApiTelemetry.report_api_use')
        self.mock_report_api_use = patcher.start()
        self.addCleanup(patcher.stop)

    def test_session_id(self):
        """
        Session ID stays the same across the calls.
        """
        test_api_function1()
        self.mock_report_api_use.assert_called_once_with(mock.sentinel.API_NAME1, mock.ANY)
        session_id = self.mock_report_api_use.call_args[0][1]

        self.mock_report_api_use.reset_mock()
        test_api_function2()
        self.mock_report_api_use.assert_called_once_with(mock.sentinel.API_NAME2, session_id)

    def test_session_reset_session_id(self):
        test_api_function1()
        self.mock_report_api_use.assert_called_once_with(mock.sentinel.API_NAME1, mock.ANY)
        session_id = self.mock_report_api_use.call_args[0][1]

        self.mock_report_api_use.reset_mock()
        quilt3.telemetry.reset_session_id()
        test_api_function2()
        self.mock_report_api_use.assert_called_once_with(mock.sentinel.API_NAME2, mock.ANY)
        new_session_id = self.mock_report_api_use.call_args[0][1]

        assert new_session_id != session_id



from quilt3.util import PhysicalKey, fix_url, get_from_config

from . import local, s3
from .base import PackageRegistry


def get_package_registry(path=None) -> PackageRegistry:
    """ Returns the package registry for a given path """
    # TODO: Don't check if it's PackageRegistry? Then we need better separation
    #       to external functions that receive string and internal that receive
    #       PackageRegistry.
    if isinstance(path, PackageRegistry):
        return path
    if not isinstance(path, PhysicalKey):
        path = PhysicalKey.from_url(
            get_from_config('default_local_registry') if path is None else fix_url(path)
        )
    return (local if path.is_local() else s3).get_package_registry(
        version=int(get_from_config('default_registry_version')),
    )(path)



from .backends import get_package_registry
from .data_transfer import copy_file
from .search_util import search_api
from .telemetry import ApiTelemetry
from .util import (
    CONFIG_PATH,
    CONFIG_TEMPLATE,
    PhysicalKey,
    QuiltConfig,
    QuiltException,
    config_exists,
    configure_from_default,
    configure_from_url,
    fix_url,
    load_config,
    read_yaml,
    validate_package_name,
    write_yaml,
)


def copy(src, dest):
    """
    Copies ``src`` object from QUILT to ``dest``.

    Either of ``src`` and ``dest`` may be S3 paths (starting with ``s3://``)
    or local file paths (starting with ``file:///``).

    Parameters:
        src (str): a path to retrieve
        dest (str): a path to write to
    """
    copy_file(PhysicalKey.from_url(fix_url(src)), PhysicalKey.from_url(fix_url(dest)))


@ApiTelemetry("api.delete_package")
def delete_package(name, registry=None, top_hash=None):
    """
    Delete a package. Deletes only the manifest entries and not the underlying files.

    Parameters:
        name (str): Name of the package
        registry (str): The registry the package will be removed from
        top_hash (str): Optional. A package hash to delete, instead of the whole package.
    """
    validate_package_name(name)
    registry = get_package_registry(registry)
    if top_hash is None:
        registry.delete_package(name)
    else:
        registry.delete_package_version(name, registry.resolve_top_hash(name, top_hash))


@ApiTelemetry("api.list_packages")
def list_packages(registry=None):
    """Lists Packages in the registry.

    Returns an iterable of all named packages in a registry.
    If the registry is None, default to the local registry.

    Args:
        registry (str): location of registry to load package from.

    Returns:
        An iterable of strings containing the names of the packages
    """
    return get_package_registry(registry).list_packages()


@ApiTelemetry("api.list_package_versions")
def list_package_versions(name, registry=None):
    """Lists versions of a given package.

    Returns an iterable of (version, hash) of a package in a registry.
    If the registry is None, default to the local registry.

    Args:
        name (str): Name of the package
        registry (str): location of registry to load package from.

    Returns:
        An iterable of tuples containing the version and hash for the package.
    """
    validate_package_name(name)
    return get_package_registry(registry).list_package_versions(name)


@ApiTelemetry("api.config")
def config(*catalog_url, **config_values):
    """Set or read the QUILT configuration.

    To retrieve the current config, call directly, without arguments:

        import quilt3
        quilt3.config()

    To trigger autoconfiguration, call with just the navigator URL:

        import quilt3
        quilt3.config(navigator_url='https://example.com')

    To set config values, call with one or more key=value pairs:

        import quilt3
        quilt3.config(navigator_url='http://example.com', elastic_search_url='http://example.com/queries')

    Default config values can be found in `quilt3.util.CONFIG_TEMPLATE`.

    Args:
        catalog_url: A (single) URL indicating a location to configure from
        **config_values: `key=value` pairs to set in the config

    Returns:
        QuiltConfig: (an ordered Mapping)
    """
    return _config(*catalog_url, **config_values)


def _config(*catalog_url, **config_values):
    """   telemetry-free version of config()   """
    if catalog_url and config_values:
        raise QuiltException("Expected either an auto-config URL or key=value pairs, but got both.")
    # Total distinction of args and kwargs -- config(catalog_url='http://foo.com')
    if catalog_url and len(catalog_url) > 1:
        raise QuiltException("`catalog_url` cannot be used with other `config_values`.")

    # Use given catalog's config to replace local configuration
    if catalog_url:
        catalog_url = catalog_url[0]

        # If catalog_url is empty, reset to an empty config.
        if catalog_url:
            config_template = configure_from_url(catalog_url)
        else:
            config_template = read_yaml(CONFIG_TEMPLATE)
            write_yaml(config_template, CONFIG_PATH, keep_backup=True)
        local_config = config_template
    # Create a custom config with the passed-in values only
    elif config_values:
        local_config = load_config()
        config_values = QuiltConfig('', config_values)  # Does some validation/scrubbing
        for key, value in config_values.items():
            local_config[key] = value
        write_yaml(local_config, CONFIG_PATH)
    # Return the current config if present or create one from the default stack
    else:
        if config_exists():
            local_config = load_config()
        else:
            local_config = configure_from_default()

    # Return current config
    return QuiltConfig(CONFIG_PATH, local_config)


@ApiTelemetry("api.disable_telemetry")
def disable_telemetry():
    """ Permanently disable sending of anonymous usage metrics """
    _disable_telemetry()


def _disable_telemetry():
    _config(telemetry_disabled=True)


@ApiTelemetry("api.search")
def search(query, limit=10):
    """
    Execute a search against the configured search endpoint.

    Args:
        query (str): query string to search
        limit (number): maximum number of results to return. Defaults to 10

    Query Syntax:
        [simple query string query](
            https://www.elastic.co/guide/en/elasticsearch/reference/6.8/query-dsl-simple-query-string-query.html)


    Returns:
        a list of objects with the following structure:
        ```
        [{
            "_id": <document unique id>
            "_index": <source index>,
            "_score": <relevance score>
            "_source":
                "key": <key of the object>,
                "size": <size of object in bytes>,
                "user_meta": <user metadata from meta= via quilt3>,
                "last_modified": <timestamp from ElasticSearch>,
                "updated": <object timestamp from S3>,
                "version_id": <version_id of object version>
            "_type": <document type>
        }, ...]
        ```
    """
    # force a call to configure_from_default if no config exists
    _config()
    raw_results = search_api(query, '*', limit)
    return raw_results['hits']['hits']



import contextlib
import functools
import gc
import hashlib
import io
import json
import logging
import os
import pathlib
import shutil
import sys
import tempfile
import textwrap
import time
import uuid
import warnings
from collections import deque
from multiprocessing import Pool

import botocore.exceptions
import jsonlines
from tqdm import tqdm

from . import util, workflows
from .backends import get_package_registry
from .data_transfer import (
    calculate_sha256,
    copy_file,
    copy_file_list,
    get_bytes,
    get_size_and_version,
    list_object_versions,
    list_url,
    put_bytes,
)
from .exceptions import PackageException
from .formats import CompressionRegistry, FormatRegistry
from .telemetry import ApiTelemetry
from .util import CACHE_PATH, DISABLE_TQDM, PACKAGE_UPDATE_POLICY
from .util import TEMPFILE_DIR_PATH as APP_DIR_TEMPFILE_DIR
from .util import (
    PhysicalKey,
    QuiltConflictException,
    QuiltException,
    catalog_package_url,
    extract_file_extension,
    fix_url,
    get_from_config,
    get_install_location,
    quiltignore_filter,
    user_is_configured_to_custom_stack,
    validate_key,
    validate_package_name,
)

logger = logging.getLogger(__name__)


# S3 Select limitation:
# https://docs.aws.amazon.com/AmazonS3/latest/userguide/selecting-content-from-objects.html#selecting-content-from-objects-requirements-and-limits
# > The maximum length of a record in the input or result is 1 MB.
# The actual limit is 1 MiB, but it's rounded because column names are included in this limit.
DEFAULT_MANIFEST_MAX_RECORD_SIZE = 1_000_000
MANIFEST_MAX_RECORD_SIZE = util.get_pos_int_from_env('QUILT_MANIFEST_MAX_RECORD_SIZE')
if MANIFEST_MAX_RECORD_SIZE is None:
    MANIFEST_MAX_RECORD_SIZE = DEFAULT_MANIFEST_MAX_RECORD_SIZE

SUPPORTED_HASH_TYPES = (
    "SHA256",
)


def _fix_docstring(**kwargs):
    def f(wrapped):
        if sys.flags.optimize < 2:
            wrapped.__doc__ = textwrap.dedent(wrapped.__doc__) % kwargs
        return wrapped
    return f


_WORKFLOW_PARAM_DOCSTRING = (
    'workflow: workflow ID or `None` to skip workflow validation.\n'
    '        If not specified, the default workflow will be used.\n'
    '        For details see: https://docs.quiltdata.com/advanced-usage/workflows\n'
)


def _delete_local_physical_key(pk):
    assert pk.is_local(), "This function only works on files that live on a local disk"
    pathlib.Path(pk.path).unlink()


def _filesystem_safe_encode(key):
    """Returns the sha256 of the key. This ensures there are no slashes, uppercase/lowercase conflicts,
    avoids `OSError: [Errno 36] File name too long:`, etc."""
    return hashlib.sha256(key.encode()).hexdigest()


def _check_hash_type_support(hash_type: str) -> None:
    if hash_type not in SUPPORTED_HASH_TYPES:
        raise QuiltException(
            f"Unsupported hash type: {hash_type!r}. "
            f"Supported types: {', '.join(SUPPORTED_HASH_TYPES)}. "
            "Try to update quilt3."
        )


class ObjectPathCache:
    @classmethod
    def _cache_path(cls, url):
        url_hash = _filesystem_safe_encode(url)
        return CACHE_PATH / url_hash[0:2] / url_hash[2:]

    @classmethod
    def get(cls, url):
        cache_path = cls._cache_path(url)
        try:
            with open(cache_path, encoding='utf-8') as fd:
                path, dev, ino, mtime = json.load(fd)
        except (FileNotFoundError, ValueError):
            return None

        try:
            stat = pathlib.Path(path).stat()
        except FileNotFoundError:
            return None

        # check if device, file, and timestamp are unchanged => cache hit
        # see also https://docs.python.org/3/library/os.html#os.stat_result
        if stat.st_dev == dev and stat.st_ino == ino and stat.st_mtime_ns == mtime:
            return path
        else:
            return None

    @classmethod
    def set(cls, url, path):
        stat = pathlib.Path(path).stat()
        cache_path = cls._cache_path(url)
        cache_path.parent.mkdir(parents=True, exist_ok=True)
        with open(cache_path, 'w', encoding='utf-8') as fd:
            json.dump([path, stat.st_dev, stat.st_ino, stat.st_mtime_ns], fd)

    @classmethod
    def clear(cls):
        shutil.rmtree(CACHE_PATH)


class PackageEntry:
    """
    Represents an entry at a logical key inside a package.
    """
    __slots__ = ('physical_key', 'size', 'hash', '_meta')

    def __init__(self, physical_key, size, hash_obj, meta):
        """
        Creates an entry.

        Args:
            physical_key: a URI (either `s3://` or `file://`)
            size(number): size of object in bytes
            hash({'type': string, 'value': string}): hash object
                for example: {'type': 'SHA256', 'value': 'bb08a...'}
            meta(dict): metadata dictionary

        Returns:
            a PackageEntry
        """
        assert isinstance(physical_key, PhysicalKey)
        self.physical_key = physical_key
        self.size = size
        self.hash = hash_obj
        self._meta = meta or {}

    def __eq__(self, other):
        return (
            # Don't check physical keys.
            self.size == other.size
            and self.hash == other.hash
            and self._meta == other._meta
        )

    def __repr__(self):
        return f"PackageEntry('{self.physical_key}')"

    def as_dict(self):
        """
        Returns dict representation of entry.
        """
        return {
            'physical_keys': [str(self.physical_key)],
            'size': self.size,
            'hash': self.hash,
            'meta': self._meta
        }

    @property
    def meta(self):
        return self._meta.get('user_meta', {})

    def set_meta(self, meta):
        """
        Sets the user_meta for this PackageEntry.
        """
        self._meta['user_meta'] = meta

    def _verify_hash(self, read_bytes):
        """
        Verifies hash of bytes
        """
        if self.hash is None:
            raise QuiltException("Hash missing - need to build the package")
        _check_hash_type_support(self.hash.get('type'))
        digest = hashlib.sha256(read_bytes).hexdigest()
        if digest != self.hash.get('value'):
            raise QuiltException("Hash validation failed")

    def set(self, path=None, meta=None):
        """
        Returns self with the physical key set to path.

        Args:
            path(string): new path to place at logical_key in the package
                Currently only supports a path on local disk
            meta(dict): metadata dict to attach to entry. If meta is provided, set just
                updates the meta attached to logical_key without changing anything
                else in the entry

        Returns:
            self
        """
        if path is not None:
            self.physical_key = PhysicalKey.from_url(fix_url(path))
            self.size = None
            self.hash = None
        elif meta is not None:
            self.set_meta(meta)
        else:
            raise PackageException('Must specify either path or meta')

    def get(self):
        """
        Returns the physical key of this PackageEntry.
        """
        return str(self.physical_key)

    def get_cached_path(self):
        """
        Returns a locally cached physical key, if available.
        """
        if util.IS_CACHE_ENABLED and not self.physical_key.is_local():
            return ObjectPathCache.get(str(self.physical_key))
        return None

    def get_bytes(self, use_cache_if_available=True):
        """
        Returns the bytes of the object this entry corresponds to. If 'use_cache_if_available'=True, will first try to
        retrieve the bytes from cache.
        """
        if use_cache_if_available:
            cached_path = self.get_cached_path()
            if cached_path is not None:
                return get_bytes(PhysicalKey(None, cached_path, None))

        data = get_bytes(self.physical_key)
        return data

    def get_as_json(self, use_cache_if_available=True):
        """
        Returns a JSON file as a `dict`. Assumes that the file is encoded using utf-8.

        If 'use_cache_if_available'=True, will first try to retrieve the object from cache.
        """
        obj_bytes = self.get_bytes(use_cache_if_available=use_cache_if_available)
        return json.loads(obj_bytes.decode("utf-8"))

    def get_as_string(self, use_cache_if_available=True):
        """
        Return the object as a string. Assumes that the file is encoded using utf-8.

        If 'use_cache_if_available'=True, will first try to retrieve the object from cache.
        """
        obj_bytes = self.get_bytes(use_cache_if_available=use_cache_if_available)
        return obj_bytes.decode("utf-8")

    def deserialize(self, func=None, **format_opts):
        """
        Returns the object this entry corresponds to.

        Args:
            func: Skip normal deserialization process, and call func(bytes),
                returning the result directly.
            **format_opts: Some data formats may take options.  Though
                normally handled by metadata, these can be overridden here.

        Returns:
            The deserialized object from the logical_key

        Raises:
            physical key failure
            hash verification fail
            when deserialization metadata is not present
        """
        data = get_bytes(self.physical_key)

        if func is not None:
            return func(data)

        suffixes = pathlib.PurePosixPath(self.physical_key.path).suffixes

        pkey_ext = suffixes.pop() if suffixes else ''
        compression_handler = CompressionRegistry.search(pkey_ext)

        if compression_handler is not None:
            pkey_ext = suffixes.pop() if suffixes else ''

        # Verify format can be handled before checking hash.  Raises if none found.
        formats = FormatRegistry.search(None, self._meta, pkey_ext)

        # Verify hash before deserializing..
        self._verify_hash(data)

        if compression_handler is not None:
            data = compression_handler.decompress(data)

        return formats[0].deserialize(data, self._meta, pkey_ext, **format_opts)

    def fetch(self, dest=None):
        """
        Gets objects from entry and saves them to dest.

        Args:
            dest: where to put the files
                Defaults to the entry name

        Returns:
            None
        """
        if dest is None:
            name = self.physical_key.basename()
            dest = PhysicalKey.from_path('.').join(name)
        else:
            dest = PhysicalKey.from_url(fix_url(dest))

        copy_file(self.physical_key, dest)

        # return a package reroot package physical keys after the copy operation succeeds
        # see GH#388 for context
        return self.with_physical_key(dest)

    def __call__(self, func=None, **kwargs):
        """
        Shorthand for self.deserialize()
        """
        return self.deserialize(func=func, **kwargs)

    def with_physical_key(self, key):
        return self.__class__(key, self.size, self.hash, self._meta)


class PackageRevInfo:
    __slots__ = ('registry', 'name', 'top_hash')

    def __init__(self, registry, name, top_hash):
        self.registry = registry
        self.name = name
        self.top_hash = top_hash


class ManifestJSONDecoder(json.JSONDecoder):
    """
    Standard json.JSONDecoder reuses same `str` objects for JSON properties, while doing
    a single `decode()` call.
    This class also reuses `str` between many `decode()`s.
    """
    def __init__(self, *args, **kwargs):
        @functools.lru_cache(maxsize=None)
        def memoize_key(s):
            return s

        def object_pairs_hook(items):
            return {
                memoize_key(k): v
                for k, v in items
            }

        super().__init__(*args, object_pairs_hook=object_pairs_hook, **kwargs)


class Package:
    """ In-memory representation of a package """

    def __init__(self):
        self._children = {}
        self._meta = {'version': 'v0'}
        self._origin = None

    @ApiTelemetry("package.__repr__")
    def __repr__(self, max_lines=20):
        """
        String representation of the Package.
        """
        def _create_str(results_dict, level=0, parent=True):
            """
            Creates a string from the results dict
            """
            result = ''
            keys = sorted(results_dict.keys())
            if not keys:
                return result

            if parent:
                has_remote_entries = any(
                    self._map(
                        lambda lk, entry: not entry.physical_key.is_local()
                    )
                )
                pkg_type = 'remote' if has_remote_entries else 'local'
                result = f'({pkg_type} Package)\n'

            for key in keys:
                result += ' ' + ('  ' * level) + '└─' + key + '\n'
                result += _create_str(results_dict[key], level + 1, parent=False)

            return result

        if not self.keys():
            return '(empty Package)'

        # traverse the tree of package directories and entries to get the list of
        # display objects. candidates is a deque of shape
        # ((logical_key, Package | PackageEntry), [list of parent key])
        candidates = deque(([x, []] for x in self._children.items()))
        results_dict = {}
        results_total = 0
        more_objects_than_lines = False

        while candidates:
            [[logical_key, entry], parent_keys] = candidates.popleft()
            if isinstance(entry, Package):
                logical_key = logical_key + '/'
                new_parent_keys = parent_keys.copy()
                new_parent_keys.append(logical_key)
                for child_key in sorted(entry.keys()):
                    candidates.append([[child_key, entry[child_key]], new_parent_keys])

            current_result_level = results_dict
            for key in parent_keys:
                current_result_level = current_result_level[key]
            current_result_level[logical_key] = {}
            results_total += 1

            if results_total >= max_lines:
                more_objects_than_lines = True
                break

        repr_str = _create_str(results_dict)

        # append '...' if the package is larger than max_size
        if more_objects_than_lines:
            repr_str += ' ' + '...\n'

        return repr_str

    @property
    def meta(self):
        return self._meta.get('user_meta', {})

    @classmethod
    @ApiTelemetry("package.install")
    def install(cls, name, registry=None, top_hash=None, dest=None, dest_registry=None, *, path=None):
        """
        Installs a named package to the local registry and downloads its files.

        Args:
            name(str): Name of package to install.
            registry(str): Registry where package is located.
                Defaults to the default remote registry.
            top_hash(str): Hash of package to install. Defaults to latest.
            dest(str): Local path to download files to.
            dest_registry(str): Registry to install package to. Defaults to local registry.
            path(str): If specified, downloads only `path` or its children.
        """
        if registry is None:
            registry = get_from_config('default_remote_registry')
            if registry is None:
                raise QuiltException(
                    "No registry specified and no default_remote_registry configured. Please "
                    "specify a registry or configure a default remote registry with quilt3.config"
                )
        else:
            registry = fix_url(registry)
        dest_registry = get_package_registry(dest_registry)
        if not dest_registry.is_local:
            raise QuiltException(
                f"Can only 'install' to a local registry, but 'dest_registry' "
                f"{dest_registry!r} is a remote path. To store a package in a remote "
                f"registry, use 'push' or 'build' instead."
            )

        if dest is None:
            dest_parsed = PhysicalKey.from_url(get_install_location()).join(name)
        else:
            dest_parsed = PhysicalKey.from_url(fix_url(dest))
            if not dest_parsed.is_local():
                raise QuiltException(
                    f"Invalid package destination path {dest!r}. 'dest', if set, must point at "
                    f"the local filesystem. To copy a package to a remote registry use 'push' or "
                    f"'build' instead."
                )

        if path:
            validate_key(path)
            subpkg_key = path
        else:
            subpkg_key = None

        pkg = cls._browse(name=name, registry=registry, top_hash=top_hash)
        message = pkg._meta.get('message', None)  # propagate the package message

        file_list = []

        if subpkg_key is not None:
            if subpkg_key not in pkg:
                raise QuiltException(f"Package {name} doesn't contain {subpkg_key!r}.")
            entry = pkg[subpkg_key]
            entries = entry.walk() if isinstance(entry, Package) else ((subpkg_key.split('/')[-1], entry),)
        else:
            entries = pkg.walk()
        for logical_key, entry in entries:
            # Copy the datafiles in the package.
            physical_key = entry.physical_key

            if util.IS_CACHE_ENABLED:
                # Try a local cache.
                cached_file = ObjectPathCache.get(str(physical_key))
                if cached_file is not None:
                    physical_key = PhysicalKey.from_path(cached_file)

            new_physical_key = dest_parsed.join(logical_key)
            if physical_key != new_physical_key:
                file_list.append((physical_key, new_physical_key, entry.size))

        def _maybe_add_to_cache(old: PhysicalKey, new: PhysicalKey, _):
            if not old.is_local() and new.is_local():
                ObjectPathCache.set(str(old), new.path)

        copy_file_list(
            file_list,
            callback=_maybe_add_to_cache if util.IS_CACHE_ENABLED else None,
            message="Copying objects",
        )

        pkg._build(name, registry=dest_registry, message=message)
        if top_hash is None:
            top_hash = pkg.top_hash
        short_top_hash = dest_registry.shorten_top_hash(name, top_hash)
        print(f"Successfully installed package '{name}', tophash={short_top_hash} from {registry}")

    @classmethod
    def resolve_hash(cls, name, registry, hash_prefix):
        """
        Find a hash that starts with a given prefix.

        Args:
            name (str): name of package
            registry (str): location of registry
            hash_prefix (str): hash prefix with length between 6 and 64 characters
        """
        validate_package_name(name)
        return get_package_registry(registry).resolve_top_hash(name, hash_prefix)

    @classmethod
    @ApiTelemetry("package.browse")
    def browse(cls, name, registry=None, top_hash=None):
        """
        Load a package into memory from a registry without making a local copy of
        the manifest.

        Args:
            name(string): name of package to load
            registry(string): location of registry to load package from
            top_hash(string): top hash of package version to load
        """
        return cls._browse(name=name, registry=registry, top_hash=top_hash)

    @classmethod
    def _browse(cls, name, registry=None, top_hash=None):
        validate_package_name(name)
        registry = get_package_registry(registry)

        top_hash = (
            get_bytes(registry.pointer_latest_pk(name)).decode()
            if top_hash is None else
            registry.resolve_top_hash(name, top_hash)
        )
        pkg_manifest = registry.manifest_pk(name, top_hash)

        def download_manifest(dst):
            copy_file(pkg_manifest, PhysicalKey.from_path(dst), message="Downloading manifest")

        with contextlib.ExitStack() as stack:
            if pkg_manifest.is_local():
                local_pkg_manifest = pkg_manifest.path
            elif util.IS_CACHE_ENABLED:
                local_pkg_manifest = CACHE_PATH / "manifest" / _filesystem_safe_encode(str(pkg_manifest))
                if not local_pkg_manifest.exists():
                    # Copy to a temporary file first, to make sure we don't cache a truncated file
                    # if the download gets interrupted.
                    tmp_path = local_pkg_manifest.with_suffix('.tmp')
                    download_manifest(tmp_path)
                    tmp_path.rename(local_pkg_manifest)
            else:
                # This tmp file has to closed before downloading, because on Windows it can't be
                # opened for concurrent access.
                with tempfile.NamedTemporaryFile(delete=False) as tmp_file:
                    local_pkg_manifest = tmp_file.name
                    stack.callback(os.unlink, local_pkg_manifest)
                download_manifest(local_pkg_manifest)

            pkg = cls._from_path(local_pkg_manifest)
            pkg._origin = PackageRevInfo(str(registry.base), name, top_hash)
            return pkg

    @classmethod
    def _from_path(cls, path):
        """ Takes a path and returns a package loaded from that path"""
        with open(path, encoding='utf-8') as open_file:
            pkg = cls._load(open_file)
        return pkg

    @classmethod
    def _split_key(cls, logical_key):
        """
        Converts a string logical key like 'a/b/c' into a list of ['a', 'b', 'c'].
        Returns the original key if it's already a list or a tuple.
        """
        if isinstance(logical_key, str):
            path = logical_key.split('/')
        elif isinstance(logical_key, (tuple, list)):
            path = logical_key
        else:
            raise TypeError('Invalid logical_key: %r' % logical_key)
        return path

    def __contains__(self, logical_key):
        """
        Checks whether the package contains a specified logical_key.

        Returns:
            True or False
        """
        try:
            self[logical_key]
            return True
        except KeyError:
            return False

    def __getitem__(self, logical_key):
        """
        Filters the package based on prefix, and returns either a new Package
            or a PackageEntry.

        Args:
            prefix(str): prefix to filter on

        Returns:
            PackageEntry if prefix matches a logical_key exactly
            otherwise Package
        """
        pkg = self
        for key_fragment in self._split_key(logical_key):
            pkg = pkg._children[key_fragment]
        return pkg

    @ApiTelemetry("package.fetch")
    def fetch(self, dest='./'):
        """
        Copy all descendants to `dest`. Descendants are written under their logical
        names _relative_ to self.

        Args:
            dest: where to put the files (locally)

        Returns:
            A new Package object with entries from self, but with physical keys
                pointing to files in `dest`.
        """
        nice_dest = PhysicalKey.from_url(fix_url(dest))
        file_list = []
        pkg = Package()

        for logical_key, entry in self.walk():
            physical_key = entry.physical_key
            new_physical_key = nice_dest.join(logical_key)

            file_list.append((physical_key, new_physical_key, entry.size))

            # return a package reroot package physical keys after the copy operation succeeds
            # see GH#388 for context
            new_entry = entry.with_physical_key(new_physical_key)
            pkg._set(logical_key, new_entry)

        copy_file_list(file_list, message="Copying objects")

        return pkg

    def keys(self):
        """
        Returns logical keys in the package.
        """
        return self._children.keys()

    def __iter__(self):
        return iter(self._children)

    def __len__(self):
        return len(self._children)

    def walk(self):
        """
        Generator that traverses all entries in the package tree and returns tuples of (key, entry),
        with keys in alphabetical order.
        """
        for name, child in sorted(self._children.items()):
            if isinstance(child, PackageEntry):
                yield name, child
            else:
                yield from child._walk(f'{name}/')

    def _walk(self, prefix):
        for name, child in sorted(self._children.items()):
            if isinstance(child, PackageEntry):
                yield f'{prefix}{name}', child
            else:
                yield from child._walk(f'{prefix}{name}/')

    def _walk_dir_meta(self):
        """
        Generator that traverses all entries in the package tree and returns
            tuples of (key, meta) for each directory with metadata.
        Keys will all end in '/' to indicate that they are directories.
        """
        for key, child in sorted(self._children.items()):
            if isinstance(child, PackageEntry):
                continue
            meta = child.meta
            if meta:
                yield key + '/', meta
            for child_key, child_meta in child._walk_dir_meta():
                yield key + '/' + child_key, child_meta

    @classmethod
    @ApiTelemetry("package.load")
    def load(cls, readable_file):
        """
        Loads a package from a readable file-like object.

        Args:
            readable_file: readable file-like object to deserialize package from

        Returns:
            A new Package object

        Raises:
            file not found
            json decode error
            invalid package exception
        """
        return cls._load(readable_file=readable_file)

    @classmethod
    def _load(cls, readable_file):
        gc.disable()  # Experiments with COCO (650MB manifest) show disabling GC gives us ~2x performance improvement

        try:
            line_count = 0
            for _ in readable_file:
                line_count += 1
            readable_file.seek(0)

            reader = jsonlines.Reader(
                tqdm(
                    readable_file,
                    desc="Loading manifest",
                    total=line_count,
                    unit="",
                    unit_scale=True,
                    disable=DISABLE_TQDM,
                    bar_format='{l_bar}{bar}| {n}/{total} [{elapsed}<{remaining}, {rate_fmt}]',
                ),
                loads=ManifestJSONDecoder().decode,
            )
            meta = reader.read()
            meta.pop('top_hash', None)  # Obsolete as of PR #130
            pkg = cls()
            pkg._meta = meta

            for obj in reader:
                path = cls._split_key(obj.pop('logical_key'))
                subpkg = pkg._ensure_subpackage(path[:-1])
                key = path[-1]
                if not obj.get('physical_keys', None):
                    # directory-level metadata
                    subpkg.set_meta(obj['meta'])
                    continue
                if key in subpkg._children:
                    raise PackageException("Duplicate logical key while loading package")
                subpkg._children[key] = PackageEntry(
                    PhysicalKey.from_url(obj['physical_keys'][0]),
                    obj['size'],
                    obj['hash'],
                    obj['meta'],
                )
        finally:
            gc.enable()
        return pkg

    def set_dir(self, lkey, path=None, meta=None, update_policy="incoming"):
        """
        Adds all files from `path` to the package.

        Recursively enumerates every file in `path`, and adds them to
            the package according to their relative location to `path`.

        Args:
            lkey(string): prefix to add to every logical key,
                use '/' for the root of the package.
            path(string): path to scan for files to add to package.
                If None, lkey will be substituted in as the path.
            meta(dict): user level metadata dict to attach to lkey directory entry.
            update_policy(str): can be either 'incoming' (default) or 'existing'.
                If 'incoming', whenever logical keys match, always take the new entry from set_dir.
                If 'existing', whenever logical keys match, retain existing entries
                and ignore new entries from set_dir.

        Returns:
            self

        Raises:
            PackageException: When `path` doesn't exist.
            ValueError: When `update_policy` is invalid.
        """
        if update_policy not in PACKAGE_UPDATE_POLICY:
            raise ValueError(f"Update policy should be one of {PACKAGE_UPDATE_POLICY}, not {update_policy!r}")

        lkey = lkey.strip("/")

        if not lkey or lkey == '.' or lkey == './':
            root = self
        else:
            validate_key(lkey)
            root = self._ensure_subpackage(self._split_key(lkey))

        root.set_meta(meta)

        if path:
            src = PhysicalKey.from_url(fix_url(path))
        else:
            src = PhysicalKey.from_path(lkey)

        # TODO: deserialization metadata
        if src.is_local():
            src_path = pathlib.Path(src.path)
            if not src_path.is_dir():
                raise PackageException("The specified directory doesn't exist")

            files = src_path.rglob('*')
            ignore = src_path / '.quiltignore'
            if ignore.exists():
                files = quiltignore_filter(files, ignore, 'file')

            for f in files:
                if not f.is_file():
                    continue
                logical_key = f.relative_to(src_path).as_posix()
                # check update policy
                if update_policy == 'existing' and logical_key in root:
                    continue
                entry = PackageEntry(PhysicalKey.from_path(f), f.stat().st_size, None, None)
                root._set(logical_key, entry)
        else:
            if src.version_id is not None:
                raise PackageException("Directories cannot have versions")
            src_path = src.path
            if src.basename() != '':
                src_path += '/'
            objects, _ = list_object_versions(src.bucket, src_path)
            for obj in objects:
                if not obj['IsLatest']:
                    continue
                # Skip S3 pseduo directory files and Keys that end in /
                if obj['Key'].endswith('/'):
                    if obj['Size'] != 0:
                        warnings.warn(f'Logical keys cannot end in "/", skipping: {obj["Key"]}')
                    continue
                obj_pk = PhysicalKey(src.bucket, obj['Key'], obj.get('VersionId'))
                logical_key = obj['Key'][len(src_path):]
                # check update policy
                if update_policy == 'existing' and logical_key in root:
                    continue
                entry = PackageEntry(obj_pk, obj['Size'], None, None)
                root._set(logical_key, entry)

        return self

    def get(self, logical_key):
        """
        Gets object from logical_key and returns its physical path.
        Equivalent to self[logical_key].get().

        Args:
            logical_key(string): logical key of the object to get

        Returns:
            Physical path as a string.

        Raises:
            KeyError: when logical_key is not present in the package
            ValueError: if the logical_key points to a Package rather than PackageEntry.
        """
        obj = self[logical_key]
        if not isinstance(obj, PackageEntry):
            raise ValueError("Key does not point to a PackageEntry")
        return obj.get()

    def readme(self):
        """
        Returns the README PackageEntry

        The README is the entry with the logical key 'README.md' (case-sensitive). Will raise a QuiltException if
        no such entry exists.
        """
        if "README.md" not in self:
            ex_msg = "This Package is missing a README file. A Quilt recognized README file is a  file named " \
                     "'README.md' (case-insensitive)"
            raise QuiltException(ex_msg)

        return self["README.md"]

    def set_meta(self, meta):
        """
        Sets user metadata on this Package.
        """
        self._meta['user_meta'] = meta
        return self

    def _fix_sha256(self):
        """
        Calculate and set missing hash values
        """
        logger.info('fix package hashes: started')

        self._incomplete_entries = [entry for key, entry in self.walk() if entry.hash is None]

        physical_keys = []
        sizes = []
        for entry in self._incomplete_entries:
            physical_keys.append(entry.physical_key)
            sizes.append(entry.size)

        results = calculate_sha256(physical_keys, sizes)
        exc = None
        for entry, obj_hash in zip(self._incomplete_entries, results):
            if isinstance(obj_hash, Exception):
                exc = obj_hash
            else:
                entry.hash = dict(type='SHA256', value=obj_hash)
        if exc:
            incomplete_manifest_path = self._dump_manifest_to_scratch()
            msg = "Unable to reach S3 for some hash values. Incomplete manifest saved to {path}."
            raise PackageException(msg.format(path=incomplete_manifest_path)) from exc

        logger.info('fix package hashes: finished')

    def _set_commit_message(self, msg):
        """
        Sets a commit message.

        Args:
            msg: a message string

        Returns:
            None

        Raises:
            a ValueError if msg is not a string
        """
        if msg is not None and not isinstance(msg, str):
            raise ValueError(
                f"The package commit message must be a string, but the message provided is an "
                f"instance of {type(msg)}."
            )

        self._meta.update({'message': msg})

    def _dump_manifest_to_scratch(self):
        registry = get_from_config('default_local_registry')
        registry_parsed = PhysicalKey.from_url(registry)
        pkg_manifest_file = registry_parsed.join("scratch").join(str(int(time.time())))

        manifest = io.BytesIO()
        self._dump(manifest)
        put_bytes(
            manifest.getvalue(),
            pkg_manifest_file
        )
        return pkg_manifest_file.path

    @property
    def _workflow(self):
        return self._meta.get('workflow')

    @_workflow.setter
    def _workflow(self, value):
        if value:
            self._meta['workflow'] = value
        else:
            self._meta.pop('workflow', None)

    def _validate_with_workflow(self, *, registry, workflow, name, message):
        self._workflow = workflows.validate(registry=registry, workflow=workflow, name=name, pkg=self, message=message)

    @ApiTelemetry("package.build")
    @_fix_docstring(workflow=_WORKFLOW_PARAM_DOCSTRING)
    def build(self, name, registry=None, message=None, *, workflow=...):
        """
        Serializes this package to a registry.

        Args:
            name: optional name for package
            registry: registry to build to
                defaults to local registry
            message: the commit message of the package
            %(workflow)s

        Returns:
            The top hash as a string.
        """
        registry = get_package_registry(registry)
        self._validate_with_workflow(registry=registry, workflow=workflow, name=name, message=message)
        return self._build(name=name, registry=registry, message=message)

    def _build(self, name, registry, message):
        validate_package_name(name)
        registry = get_package_registry(registry)

        self._set_commit_message(message)
        self._fix_sha256()

        top_hash = self.top_hash
        self._push_manifest(name, registry, top_hash)
        return top_hash

    def _push_manifest(self, name, registry, top_hash):
        manifest = io.BytesIO()
        self._dump(manifest)
        registry.push_manifest(name, top_hash, manifest.getvalue())

    @ApiTelemetry("package.dump")
    def dump(self, writable_file):
        """
        Serializes this package to a writable file-like object.

        Args:
            writable_file: file-like object to write serialized package.

        Returns:
            None

        Raises:
            fail to create file
            fail to finish write
        """
        return self._dump(writable_file)

    def _dump(self, writable_file):
        json_encode = json.JSONEncoder(ensure_ascii=False).encode

        def dumps(obj):
            data = json_encode(obj)
            encoded_size = len(data.encode())
            if encoded_size > MANIFEST_MAX_RECORD_SIZE:
                lk = obj.get('logical_key')
                entry_text = 'package metadata' if lk is None else f'entry with logical key {lk!r}'
                raise QuiltException(
                    f"Size of manifest record for {entry_text} is {encoded_size} bytes, "
                    f"but must be less than {MANIFEST_MAX_RECORD_SIZE} bytes. "
                    'Quilt recommends less than 1 MB of metadata per object, '
                    'and less than 1 MB of package-level metadata. '
                    'This enables S3 select, Athena and downstream services '
                    'to work correctly. This limit can be overridden with the '
                    'QUILT_MANIFEST_MAX_RECORD_SIZE environment variable.'
                )
            return data

        writer = jsonlines.Writer(writable_file, dumps=dumps)
        for line in self.manifest:
            writer.write(line)

    @property
    def manifest(self):
        """
        Provides a generator of the dicts that make up the serialized package.
        """
        yield self._meta
        for dir_key, meta in self._walk_dir_meta():
            yield {'logical_key': dir_key, 'meta': meta}
        for logical_key, entry in self.walk():
            yield {'logical_key': logical_key, **entry.as_dict()}

    def set(self, logical_key, entry=None, meta=None, serialization_location=None, serialization_format_opts=None):
        """
        Returns self with the object at logical_key set to entry.

        Args:
            logical_key(string): logical key to update
            entry(PackageEntry OR string OR object): new entry to place at logical_key in the package.
                If entry is a string, it is treated as a URL, and an entry is created based on it.
                If entry is None, the logical key string will be substituted as the entry value.
                If entry is an object and quilt knows how to serialize it, it will immediately be serialized and
                written to disk, either to serialization_location or to a location managed by quilt. List of types that
                Quilt can serialize is available by calling `quilt3.formats.FormatRegistry.all_supported_formats()`
            meta(dict): user level metadata dict to attach to entry
            serialization_format_opts(dict): Optional. If passed in, only used if entry is an object. Options to help
                Quilt understand how the object should be serialized. Useful for underspecified file formats like csv
                when content contains confusing characters. Will be passed as kwargs to the FormatHandler.serialize()
                function. See docstrings for individual FormatHandlers for full list of options -
                https://github.com/quiltdata/quilt/blob/master/api/python/quilt3/formats.py
            serialization_location(string): Optional. If passed in, only used if entry is an object. Where the
                serialized object should be written, e.g. "./mydataframe.parquet"

        Returns:
            self
        """
        return self._set(logical_key=logical_key,
                         entry=entry,
                         meta=meta,
                         serialization_location=serialization_location,
                         serialization_format_opts=serialization_format_opts)

    def _set(self, logical_key, entry=None, meta=None, serialization_location=None, serialization_format_opts=None):
        if not logical_key or logical_key.endswith('/'):
            raise QuiltException(
                f"Invalid logical key {logical_key!r}. "
                f"A package entry logical key cannot be a directory."
            )

        validate_key(logical_key)

        if entry is None:
            entry = pathlib.Path(logical_key).resolve().as_uri()

        if isinstance(entry, (str, os.PathLike)):
            src = PhysicalKey.from_url(fix_url(str(entry)))
            size, version_id = get_size_and_version(src)

            # Determine if a new version needs to be appended.
            if not src.is_local() and src.version_id is None and version_id is not None:
                src.version_id = version_id
            entry = PackageEntry(src, size, None, None)
        elif isinstance(entry, PackageEntry):
            assert meta is None

        elif FormatRegistry.object_is_serializable(entry):
            # Use file extension from serialization_location, fall back to file extension from logical_key
            # If neither has a file extension, Quilt picks the serialization format.
            logical_key_ext = extract_file_extension(logical_key)

            serialize_loc_ext = None
            if serialization_location is not None:
                serialize_loc_ext = extract_file_extension(serialization_location)

            if logical_key_ext is not None and serialize_loc_ext is not None:
                assert logical_key_ext == serialize_loc_ext, f"The logical_key and the serialization_location have " \
                                                             f"different file extensions: {logical_key_ext} vs " \
                                                             f"{serialize_loc_ext}. Quilt doesn't know which to use!"

            if serialize_loc_ext is not None:
                ext = serialize_loc_ext
            elif logical_key_ext is not None:
                ext = logical_key_ext
            else:
                ext = None

            format_handlers = FormatRegistry.search(type(entry))
            if ext:
                format_handlers = [f for f in format_handlers if ext in f.handled_extensions]

            if len(format_handlers) == 0:
                error_message = f'Quilt does not know how to serialize a {type(entry)}'
                if ext is not None:
                    error_message += f' as a {ext} file.'
                error_message += '. If you think this should be supported, please open an issue or PR at ' \
                                 'https://github.com/quiltdata/quilt'
                raise QuiltException(error_message)

            if serialization_format_opts is None:
                serialization_format_opts = {}
            serialized_object_bytes, new_meta = format_handlers[0].serialize(entry, meta=None, ext=ext,
                                                                             **serialization_format_opts)
            if serialization_location is None:
                serialization_path = APP_DIR_TEMPFILE_DIR / str(uuid.uuid4())
                if ext:
                    serialization_path = serialization_path.with_suffix(f'.{ext}')
            else:
                serialization_path = pathlib.Path(serialization_location).expanduser().resolve()

            serialization_path.parent.mkdir(exist_ok=True, parents=True)
            serialization_path.write_bytes(serialized_object_bytes)

            size = serialization_path.stat().st_size
            write_pk = PhysicalKey.from_path(serialization_path)
            entry = PackageEntry(write_pk, size, hash_obj=None, meta=new_meta)

        else:
            raise TypeError(f"Expected a string for entry, but got an instance of {type(entry)}.")

        if meta is not None:
            entry.set_meta(meta)

        path = self._split_key(logical_key)

        pkg = self._ensure_subpackage(path[:-1], ensure_no_entry=True)
        if path[-1] in pkg and isinstance(pkg[path[-1]], Package):
            raise QuiltException("Cannot overwrite directory with PackageEntry")
        pkg._children[path[-1]] = entry

        return self

    def _ensure_subpackage(self, path, ensure_no_entry=False):
        """
        Creates a package and any intermediate packages at the given path.

        Args:
            path(list): logical key as a list or tuple
            ensure_no_entry(boolean): if True, throws if this would overwrite
                a PackageEntry that already exists in the tree.

        Returns:
            newly created or existing package at that path
        """
        pkg = self
        for key_fragment in path:
            if ensure_no_entry and key_fragment in pkg \
                    and isinstance(pkg[key_fragment], PackageEntry):
                raise QuiltException("Already a PackageEntry along the path.")
            pkg = pkg._children.setdefault(key_fragment, Package())
        return pkg

    def delete(self, logical_key):
        """
        Returns the package with logical_key removed.

        Returns:
            self

        Raises:
            KeyError: when logical_key is not present to be deleted
        """
        path = self._split_key(logical_key)
        pkg = self[path[:-1]]
        del pkg._children[path[-1]]
        return self

    @property
    def top_hash(self):
        """
        Returns the top hash of the package.

        Note that physical keys are not hashed because the package has
            the same semantics regardless of where the bytes come from.

        Returns:
            A string that represents the top hash of the package
        """
        return self._calculate_top_hash(self._meta, self.walk())

    @classmethod
    def _calculate_top_hash(cls, meta, entries):
        top_hash = hashlib.sha256()

        json_encode = json.JSONEncoder(sort_keys=True, separators=(',', ':')).encode
        for part in cls._get_top_hash_parts(meta, entries):
            top_hash.update(json_encode(part).encode())

        return top_hash.hexdigest()

    @classmethod
    def _get_top_hash_parts(cls, meta, entries):
        assert 'top_hash' not in meta
        yield meta
        # TODO: dir-level metadata should affect top hash as well.
        for logical_key, entry in entries:
            if entry.hash is None or entry.size is None:
                raise QuiltException(
                    "PackageEntry missing hash and/or size: %s" % entry.physical_key
                )
            yield {
                'hash': entry.hash,
                'logical_key': logical_key,
                'meta': entry._meta,
                'size': entry.size,
            }

    @ApiTelemetry("package.push")
    @_fix_docstring(workflow=_WORKFLOW_PARAM_DOCSTRING)
    def push(
        self, name, registry=None, dest=None, message=None, selector_fn=None, *,
        workflow=..., force: bool = False, dedupe: bool = False
    ):
        """
        Copies objects to path, then creates a new package that points to those objects.
        Copies each object in this package to path according to logical key structure,
        then adds to the registry a serialized version of this package with
        physical keys that point to the new copies.

        Note that push is careful to not push data unnecessarily. To illustrate, imagine you have
        a PackageEntry: `pkg["entry_1"].physical_key = "/tmp/package_entry_1.json"`

        If that entry would be pushed to `s3://bucket/prefix/entry_1.json`, but
        `s3://bucket/prefix/entry_1.json` already contains the exact same bytes as
        '/tmp/package_entry_1.json', `quilt3` will not push the bytes to s3, no matter what
        `selector_fn('entry_1', pkg["entry_1"])` returns.

        However, selector_fn will dictate whether the new package points to the local file or to s3:

        If `selector_fn('entry_1', pkg["entry_1"]) == False`,
        `new_pkg["entry_1"] = ["/tmp/package_entry_1.json"]`

        If `selector_fn('entry_1', pkg["entry_1"]) == True`,
        `new_pkg["entry_1"] = ["s3://bucket/prefix/entry_1.json"]`

        By default, push will not overwrite an existing package if its top hash does not match
        the parent hash of the package being pushed. Use `force=True` to skip the check.

        Args:
            name: name for package in registry
            dest: where to copy the objects in the package. Must be either an S3 URI prefix (e.g., s3://$bucket/$key)
                in the registry bucket, or a callable that takes logical_key, package_entry, and top_hash
                and returns an S3 URI.
            registry: registry where to create the new package
            message: the commit message for the new package
            selector_fn: An optional function that determines which package entries should be copied to S3.
                The function takes in two arguments, logical_key and package_entry, and should return False if that
                PackageEntry should not be copied to the destination registry during push.
                If for example you have a package where the files are spread over multiple buckets
                and you add a single local file, you can use selector_fn to only
                push the local file to s3 (instead of pushing all data to the destination bucket).
            %(workflow)s
            force: skip the top hash check and overwrite any existing package
            dedupe: don't push if the top hash matches the existing package top hash; return the current package

        Returns:
            A new package that points to the copied objects.
        """
        return self._push(
            name, registry, dest, message, selector_fn, workflow=workflow,
            print_info=True, force=force, dedupe=dedupe
        )

    def _push(
        self, name, registry=None, dest=None, message=None, selector_fn=None, *,
        workflow, print_info, force: bool, dedupe: bool
    ):
        if selector_fn is None:
            def selector_fn(*args):
                return True

        validate_package_name(name)

        if registry is None:
            registry = get_from_config('default_remote_registry')
            if registry is None:
                raise QuiltException(
                    "No registry specified and no default remote registry configured. Please "
                    "specify a registry or configure a default remote registry with quilt3.config"
                )
            registry_parsed = PhysicalKey.from_url(fix_url(registry))
        else:
            registry_parsed = PhysicalKey.from_url(fix_url(registry))
            if not registry_parsed.is_local():
                if registry_parsed.path != '':
                    raise QuiltException(
                        f"The 'registry' argument expects an S3 bucket but the S3 object path "
                        f"{registry!r} was provided instead. You probably wanted to set "
                        f"'registry' to {'s3://' + registry_parsed.bucket!r} instead. To specify that package "
                        f"data land in a specific directory use 'dest'."
                    )
            else:
                raise QuiltException(
                    f"Can only 'push' to remote registries in S3, but {registry!r} "
                    f"is a local file. To store a package in the local registry, use "
                    f"'build' instead."
                )

        if callable(dest):
            def dest_fn(*args, **kwargs):
                url = dest(*args, **kwargs)
                if not isinstance(url, str):
                    raise TypeError(f'{dest!r} returned {url!r}, but str is expected')
                pk = PhysicalKey.from_url(url)
                if pk.is_local():
                    raise util.URLParseError("Unexpected scheme: 'file'")
                if pk.version_id:
                    raise ValueError(f'{dest!r} returned {url!r}, but URI must not include versionId')
                return pk
        else:
            def dest_fn(lk, *args, **kwargs):
                return dest_parsed.join(lk)
            if dest is None:
                dest_parsed = registry_parsed.join(name)
            else:
                dest_parsed = PhysicalKey.from_url(fix_url(dest))
                if dest_parsed.bucket != registry_parsed.bucket:
                    raise QuiltException(
                        f"Invalid package destination path {dest!r}. 'dest', if set, must be a path "
                        f"in the {registry!r} package registry specified by 'registry'."
                    )

        registry = get_package_registry(registry)
        self._validate_with_workflow(registry=registry, workflow=workflow, name=name, message=message)

        def get_latest_hash():
            try:
                return get_bytes(registry.pointer_latest_pk(name)).decode()
            except botocore.exceptions.ClientError as ex:
                if ex.response['Error']['Code'] == 'NoSuchKey':
                    # Expected
                    return None
                raise

        def check_hash_conficts(latest_hash):
            if latest_hash is None:
                return

            if self._origin is None or latest_hash != self._origin.top_hash:
                raise QuiltConflictException(
                    f"Package with hash {latest_hash} already exists at the destination; "
                    f"expected {None if self._origin is None else self._origin.top_hash}. "
                    "Use force=True (Python) or --force (CLI) to overwrite."
                )

        # Get the latest hash if we're either checking for conflicts or deduping.
        # Otherwise, avoid making unnecessary network calls.
        if not force or dedupe:
            latest_hash = get_latest_hash()
        else:
            latest_hash = None

        # Check the top hash and fail early if it's unexpected.
        if not force:
            check_hash_conficts(latest_hash)

        self._fix_sha256()

        pkg = self.__class__()
        pkg._meta = self._meta
        pkg._set_commit_message(message)
        top_hash = self._calculate_top_hash(pkg._meta, self.walk())
        pkg._origin = PackageRevInfo(str(registry.base), name, top_hash)

        if dedupe and top_hash == latest_hash:
            if print_info:
                print(
                    f"Skipping since package with hash {latest_hash} already exists "
                    "at the destination and dedupe parameter is true."
                )
            return self

        # Since all that is modified is physical keys, pkg will have the same top hash
        file_list = []
        entries = []

        for logical_key, entry in self.walk():
            if not selector_fn(logical_key, entry):
                pkg._set(logical_key, entry)
                continue

            # Copy the datafiles in the package.
            physical_key = entry.physical_key

            new_physical_key = dest_fn(logical_key, entry, top_hash)
            if (
                physical_key.bucket == new_physical_key.bucket and
                physical_key.path == new_physical_key.path
            ):
                # No need to copy - re-use the original physical key.
                pkg._set(logical_key, entry)
            else:
                entries.append((logical_key, entry))
                file_list.append((physical_key, new_physical_key, entry.size))

        results = copy_file_list(file_list, message="Copying objects")

        for (logical_key, entry), versioned_key in zip(entries, results):
            # Create a new package entry pointing to the new remote key.
            assert versioned_key is not None
            new_entry = entry.with_physical_key(versioned_key)
            pkg._set(logical_key, new_entry)

        def physical_key_is_temp_file(pk):
            if not pk.is_local():
                return False
            return pathlib.Path(pk.path).parent.resolve() == APP_DIR_TEMPFILE_DIR.resolve()

        temp_file_logical_keys = [lk for lk, entry in self.walk() if physical_key_is_temp_file(entry.physical_key)]
        if temp_file_logical_keys:
            temp_file_physical_keys = [self[lk].physical_key for lk in temp_file_logical_keys]

            # Now that data has been pushed, delete tmp files created by pkg.set('KEY', obj)
            with Pool(10) as p:
                p.map(_delete_local_physical_key, temp_file_physical_keys)

            # Update old package to point to the materialized location of the file since the tempfile no longest exists
            for lk in temp_file_logical_keys:
                self._set(lk, pkg[lk])

        # Check top hash again just before pushing, to minimize the race condition.
        if not force:
            latest_hash = get_latest_hash()
            check_hash_conficts(latest_hash)

        pkg._push_manifest(name, registry, top_hash)

        if print_info:
            shorthash = registry.shorten_top_hash(name, top_hash)
            print(f"Package {name}@{shorthash} pushed to s3://{registry.base.bucket}")

            if user_is_configured_to_custom_stack():
                navigator_url = get_from_config("navigator_url")

                print(f"Successfully pushed the new package to "
                      f"{catalog_package_url(navigator_url, registry.base.bucket, name, tree=False)}")
            else:
                dest_s3_url = str(registry.base)
                if not dest_s3_url.endswith("/"):
                    dest_s3_url += "/"
                print(f"Run `quilt3 catalog {dest_s3_url}` to browse.")
                print("Successfully pushed the new package")

        return pkg

    @classmethod
    def rollback(cls, name, registry, top_hash):
        """
        Set the "latest" version to the given hash.

        Args:
            name(str): Name of package to rollback.
            registry(str): Registry where package is located.
            top_hash(str): Hash to rollback to.
        """
        validate_package_name(name)
        registry = get_package_registry(PhysicalKey.from_url(fix_url(registry)))
        top_hash = registry.resolve_top_hash(name, top_hash)

        # Check that both latest and top_hash actually exist.
        get_size_and_version(registry.manifest_pk(name, top_hash))
        latest_path = registry.pointer_latest_pk(name)
        get_size_and_version(registry.pointer_latest_pk(name))

        put_bytes(top_hash.encode('utf-8'), latest_path)

    @ApiTelemetry("package.diff")
    def diff(self, other_pkg):
        """
        Returns three lists -- added, modified, deleted.

        Added: present in other_pkg but not in self.
        Modified: present in both, but different.
        Deleted: present in self, but not other_pkg.

        Args:
            other_pkg: Package to diff

        Returns:
            added, modified, deleted (all lists of logical keys)
        """
        deleted = []
        modified = []
        other_entries = dict(other_pkg.walk())
        for lk, entry in self.walk():
            other_entry = other_entries.pop(lk, None)
            if other_entry is None:
                deleted.append(lk)
            elif entry != other_entry:
                modified.append(lk)

        added = list(sorted(other_entries))

        return added, modified, deleted

    @ApiTelemetry("package.map")
    def map(self, f, include_directories=False):
        """
        Performs a user-specified operation on each entry in the package.

        Args:
            f(x, y): function
                The function to be applied to each package entry.
                It should take two inputs, a logical key and a PackageEntry.
            include_directories: bool
                Whether or not to include directory entries in the map.

        Returns: list
            The list of results generated by the map.
        """
        return self._map(f, include_directories=include_directories)

    def _map(self, f, include_directories=False):

        if include_directories:
            for lk, _ in self._walk_dir_meta():
                yield f(lk, self[lk.rstrip("/")])

        for lk, entity in self.walk():
            yield f(lk, entity)

    @ApiTelemetry("package.filter")
    def filter(self, f, include_directories=False):
        """
        Applies a user-specified operation to each entry in the package,
        removing results that evaluate to False from the output.

        Args:
            f(x, y): function
                The function to be applied to each package entry.
                It should take two inputs, a logical key and a PackageEntry.
                This function should return a boolean.
            include_directories: bool
                Whether or not to include directory entries in the map.

        Returns:
            A new package with entries that evaluated to False removed
        """
        return self._filter(f=f, include_directories=include_directories)

    def _filter(self, f, include_directories=False):
        p = Package()

        excluded_dirs = set()
        if include_directories:
            for lk, _ in self._walk_dir_meta():
                if not f(lk, self[lk.rstrip("/")]):
                    excluded_dirs.add(lk)

        for lk, entity in self.walk():
            if (not any(p in excluded_dirs
                        for p in pathlib.PurePosixPath(lk).parents)
                    and f(lk, entity)):
                p._set(lk, entity)

        return p

    def verify(self, src, extra_files_ok=False):
        """
        Check if the contents of the given directory matches the package manifest.

        Args:
            src(str): URL of the directory
            extra_files_ok(bool): Whether extra files in the directory should cause a failure.

        Returns:
            True if the package matches the directory; False otherwise.
        """
        for lk, e in self.walk():
            _check_hash_type_support(e.hash["type"])

        src = PhysicalKey.from_url(fix_url(src))
        src_dict = dict(list_url(src))
        url_list = []
        size_list = []
        for logical_key, entry in self.walk():
            src_size = src_dict.pop(logical_key, None)
            if src_size is None:
                return False
            if entry.size != src_size:
                return False
            entry_url = src.join(logical_key)
            url_list.append(entry_url)
            size_list.append(src_size)

        if src_dict and not extra_files_ok:
            return False

        hash_list = calculate_sha256(url_list, size_list)
        for (logical_key, entry), url_hash in zip(self.walk(), hash_list):
            if isinstance(url_hash, Exception):
                raise url_hash
            if entry.hash['value'] != url_hash:
                return False

        return True



"""
Parses the command-line arguments and runs a command.
"""

import _thread
import argparse
import functools
import json
import sys
import time

import requests

from . import Package
from . import __version__ as quilt3_version
from . import api, session, util
from .backends import get_package_registry
from .session import open_url
from .util import (
    QuiltException,
    catalog_package_url,
    catalog_s3_url,
    get_from_config,
)


def parse_arg_json(value):
    try:
        return json.loads(value)
    except json.JSONDecodeError:
        raise argparse.ArgumentTypeError(f'{value!r} is not a valid json string.')


def parse_positive_int(value):
    value = int(value)
    if value <= 0:
        raise argparse.ArgumentTypeError(f'{value!r} is not a positive integer.')
    return value


def cmd_config(catalog_url, **kwargs):
    """
    Configure quilt3 to a Quilt stack
    """
    config_values = kwargs['set'] if kwargs['set'] else {}
    if catalog_url and config_values:
        raise QuiltException("Expected either an auto-config URL or key=value pairs, but got both.")

    if config_values:
        api.config(**config_values)
    else:
        if catalog_url is None:
            existing_catalog_url = get_from_config('navigator_url')
            if existing_catalog_url is not None:
                print(existing_catalog_url)
            else:
                print('<None>')
        else:
            api.config(catalog_url)


class ParseConfigDict(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        d = {}
        if values:
            for item in values:
                split_items = item.split("=", 1)
                key, value = split_items[0].strip(), split_items[1]
                d[key] = value
        setattr(namespace, self.dest, d)


def cmd_config_default_registry(default_remote_registry):
    """
    Configure the default remote registry for quilt3
    """
    api.config(default_remote_registry=default_remote_registry)
    print(f"Successfully set the default remote registry to {default_remote_registry}")


def _test_url(url):
    try:
        response = requests.get(url)
        if response.ok:
            return True
        return False
    except requests.exceptions.ConnectionError:
        return False


catalog_cmd_detailed_help = """
Run the Quilt catalog on your machine. Running `quilt3 catalog` launches a
Python webserver on your local machine that serves a catalog web app and
provides required backend services using temporary AWS credentials.
Temporary credentials are derived from your default AWS credentials
(or active `AWS_PROFILE`) using `boto3.sts.get_session_token`.
For more details about configuring and using AWS credentials in `boto3`,
see the AWS documentation:
https://boto3.amazonaws.com/v1/documentation/api/latest/guide/configuration.html

#### Previewing files in S3
The Quilt catalog allows users to preview files in S3 by downloading and
processing/converting them inside the Python webserver running on local machine.
Neither your AWS credentials nor data requested goes through any third-party
cloud services aside of S3.
"""


def _launch_local_catalog(*, host: str, port: int):
    try:
        import uvicorn

        from quilt3_local.main import app
    except ModuleNotFoundError as e:
        if e.name in (
            'uvicorn',
            'quilt3_local',
        ):
            raise QuiltException("To run `quilt3 catalog` please `pip install 'quilt3[catalog]'`") from e
        raise
    _thread.start_new_thread(functools.partial(uvicorn.run, host=host, port=port, log_level="info"), (app,))


def cmd_catalog(*, navigation_target=None, detailed_help=False, host: str, port: int, no_browser: bool):
    """
    Run the Quilt catalog locally. If navigation_targets starts with 's3://', open file view. Otherwise assume it
    refers to a package, following the pattern: BUCKET:USER/PKG

    If detailed_help=True, display detailed information about the `quilt3 catalog` command and then exit
    """
    if detailed_help:
        print(catalog_cmd_detailed_help)
        return

    local_catalog_url = f"http://{host}:{port}"

    # Build the catalog URL - we do this at the beginning so simple syntax errors return immediately
    if navigation_target is None:
        catalog_url = local_catalog_url
    elif navigation_target.startswith("s3://"):
        catalog_url = catalog_s3_url(local_catalog_url, navigation_target)
    else:
        num_colons = navigation_target.count(":")
        assert num_colons == 1, f"To go to Package view, the input should follow the pattern BUCKET:USER/PKG. " \
            f"However the input {navigation_target} has {num_colons} colons when it should have exactly one."
        num_slashes = navigation_target.count("/")
        assert num_slashes == 1, f"To go to Package view, the input should follow the pattern BUCKET:USER/PKG. " \
            f"However the input {navigation_target} has {num_slashes} backslashes when it should have exactly one."
        bucket, package_name = navigation_target.split(":")
        catalog_url = catalog_package_url(local_catalog_url, bucket, package_name)

    should_launch_local_catalog = not _test_url(local_catalog_url)
    if should_launch_local_catalog:
        _launch_local_catalog(host=host, port=port)

        # Make sure the containers are running and available before opening the browser window
        print("Waiting for local catalog app to launch...")
        failure_timeout_secs = 15
        poll_interval_secs = 0.5
        start_time = time.time()
        while True:
            if time.time() - start_time > failure_timeout_secs:
                # Succeeded at the last second, let it proceed.
                if _test_url(local_catalog_url):
                    break
                raise QuiltException("The local catalog app did not successfully launch.")

            if _test_url(local_catalog_url):
                # Everything is working, proceed
                break
            else:
                time.sleep(poll_interval_secs)  # The containers can take a moment to launch

    if not no_browser:
        open_url(catalog_url)

    if should_launch_local_catalog:
        while True:
            try:
                time.sleep(1)
            except KeyboardInterrupt:
                sys.exit(0)


def cmd_disable_telemetry():
    api._disable_telemetry()
    print("Successfully disabled telemetry.")


def cmd_list_packages(registry):
    for package_name in get_package_registry(registry).list_packages():
        print(package_name)


def cmd_verify(name, registry, top_hash, dir, extra_files_ok):
    pkg = Package._browse(name, registry, top_hash)
    if pkg.verify(dir, extra_files_ok):
        print("Verification succeeded")
        return 0
    else:
        print("Verification failed")
        return 1


# This is not a lambda to ease of testing.
def _selector_fn_no_copy(*args):
    return False


def cmd_push(name, dir, registry, dest, message, meta, workflow, force, dedupe, no_copy):
    if util.PhysicalKey.from_url(util.fix_url(dir)).is_local() and no_copy:
        raise QuiltException("--no-copy flag can be specified only for remote data.")

    try:
        pkg = Package.browse(name, None)
    except FileNotFoundError:
        pkg = Package()

    pkg.set_dir('.', dir, meta=meta)
    pkg.push(
        name, registry=registry, dest=dest, message=message,
        workflow=workflow, force=force, dedupe=dedupe,
        **({"selector_fn": _selector_fn_no_copy} if no_copy else {}),
    )


def create_parser():
    parser = argparse.ArgumentParser(allow_abbrev=False)
    parser.add_argument(
            "--version",
            help="Show quilt3 version and exit",
            action="version",
            version=quilt3_version,
    )

    subparsers = parser.add_subparsers(metavar="<command>")
    subparsers.required = True

    # login
    shorthelp = "Log in to configured Quilt server"
    login_p = subparsers.add_parser("login", description=shorthelp, help=shorthelp, allow_abbrev=False)
    login_p.set_defaults(func=session.login)

    # logout
    shorthelp = "Log out of current Quilt server"
    logout_p = subparsers.add_parser("logout", description=shorthelp, help=shorthelp, allow_abbrev=False)
    logout_p.set_defaults(func=session.logout)

    # config
    shorthelp = "Configure Quilt"
    config_p = subparsers.add_parser("config", description=shorthelp, help=shorthelp, allow_abbrev=False)
    config_p.add_argument(
        "catalog_url",
        help="URL of catalog to config with, or empty string to reset the config",
        type=str,
        nargs="?"
    )
    config_p.add_argument(
            "--set",
            metavar="KEY=VALUE",
            nargs="+",
            help="Set a number of key-value pairs for config_values"
                 "(do not put spaces before or after the = sign). "
                 "If a value contains spaces, you should define "
                 "it with double quotes: "
                 'foo="this is a sentence". Note that '
                 "values are always treated as strings.",
            action=ParseConfigDict,
    )
    config_p.set_defaults(func=cmd_config)

    # config-default-registry
    shorthelp = "Configure default remote registry for Quilt"
    config_p = subparsers.add_parser("config-default-remote-registry",
                                     description=shorthelp, help=shorthelp, allow_abbrev=False)
    config_p.add_argument(
            "default_remote_registry",
            help="The default remote registry to use, e.g. s3://quilt-ml",
            type=str
    )
    config_p.set_defaults(func=cmd_config_default_registry)

    # catalog
    shorthelp = "Run Quilt catalog locally"
    catalog_p = subparsers.add_parser("catalog", description=shorthelp, help=shorthelp, allow_abbrev=False)
    catalog_p.add_argument(
            "navigation_target",
            help="Which page in the local catalog to open. Leave blank to go to the catalog landing page, pass in an "
                 "s3 url (e.g. 's3://bucket/myfile.txt') to go to file viewer, or pass in a package name in the form "
                 "'BUCKET:USER/PKG' to go to the package viewer.",
            type=str,
            nargs="?"
    )
    catalog_p.add_argument(
            "--detailed-help", "--detailed_help",
            help="Display detailed information about this command and then exit",
            action="store_true",
    )
    catalog_p.add_argument(
        "--host",
        type=str,
        default="127.0.0.1",
        help="Bind socket to this host",
    )
    catalog_p.add_argument(
        "--port",
        type=parse_positive_int,
        default=3000,
        help="Bind to a socket with this port",
    )
    catalog_p.add_argument(
        "--no-browser",
        action="store_true",
        help="Don't open catalog in a browser after startup",
    )
    catalog_p.set_defaults(func=cmd_catalog)

    # disable-telemetry
    shorthelp = "Disable anonymous usage metrics"
    disable_telemetry_p = subparsers.add_parser("disable-telemetry",
                                                description=shorthelp, help=shorthelp, allow_abbrev=False)
    disable_telemetry_p.set_defaults(func=cmd_disable_telemetry)

    # install
    shorthelp = "Install a package"
    install_p = subparsers.add_parser("install", description=shorthelp, help=shorthelp, allow_abbrev=False)
    install_p.add_argument(
        "name",
        help=(
            "Name of package, in the USER/PKG format"
        ),
        type=str,
    )
    install_p.add_argument(
        "--registry",
        help="Registry where package is located, usually s3://MY-BUCKET. Defaults to the default remote registry.",
        type=str,
        required=False,
    )
    install_p.add_argument(
        "--top-hash",
        help="Hash of package to install. Defaults to latest.",
        type=str,
        required=False,
    )
    install_p.add_argument(
        "--dest",
        help="Local path to download files to.",
        type=str,
        required=False,
    )
    install_p.add_argument(
        "--dest-registry",
        help="Registry to install package to. Defaults to local registry.",
        type=str,
        required=False,
    )
    install_p.add_argument(
        "--path",
        help="If specified, downloads only PATH or its children.",
        type=str,
        required=False,
    )
    install_p.set_defaults(func=Package.install)

    # list-packages
    shorthelp = "List all packages in a registry"
    list_packages_p = subparsers.add_parser("list-packages", description=shorthelp, help=shorthelp, allow_abbrev=False)
    list_packages_p.add_argument(
            "registry",
            help="Registry for packages, e.g. s3://quilt-example",
            type=str,
    )
    list_packages_p.set_defaults(func=cmd_list_packages)

    # verify
    shorthelp = "Verify that package contents matches a given directory"
    verify_p = subparsers.add_parser("verify", description=shorthelp, help=shorthelp, allow_abbrev=False)
    verify_p.add_argument(
        "name",
        help="Name of package, in the USER/PKG format",
        type=str,
    )
    verify_p.add_argument(
        "--registry",
        help="Registry where package is located, usually s3://MY-BUCKET",
        type=str,
        required=True,
    )
    verify_p.add_argument(
        "--top-hash",
        help="Hash of package to verify",
        type=str,
        required=True,
    )
    verify_p.add_argument(
        "--dir",
        help="Directory to verify",
        type=str,
        required=True,
    )
    verify_p.add_argument(
        "--extra-files-ok",
        help="Whether extra files in the directory should cause a failure",
        action="store_true"
    )
    verify_p.set_defaults(func=cmd_verify)

    # push
    shorthelp = "Pushes the new package to the remote registry"
    push_p = subparsers.add_parser("push", description=shorthelp, help=shorthelp, allow_abbrev=False, add_help=False)
    required_args = push_p.add_argument_group('required arguments')
    optional_args = push_p.add_argument_group('optional arguments')
    push_p.add_argument(
        "name",
        help="Name of package, in the USER/PKG format",
        type=str,
    )
    required_args.add_argument(
        "--dir",
        help="Directory to add to the new package",
        type=str,
        required=True,
    )
    optional_args.add_argument(
        '-h',
        '--help',
        action='help',
        default=argparse.SUPPRESS,
        help='show this help message and exit'
    )
    optional_args.add_argument(
        "--registry",
        help="Registry where to create the new package. Defaults to the default remote registry.",
        type=str,
    )
    optional_args.add_argument(
        "--dest",
        help="Where to copy the objects in the package",
        type=str,
    )
    optional_args.add_argument(
        "--message",
        help="The commit message for the new package",
        type=str,
    )
    optional_args.add_argument(
        "--meta",
        help="""
            Sets package-level metadata.
            Format: A json string with keys in double quotes '{"key": "value"}'
            """,
        type=parse_arg_json,
    )
    optional_args.add_argument(
        "--workflow",
        help="""
            Workflow ID or empty string to skip workflow validation.
            If not specified, the default workflow will be used.
            """,
        default=...,
        type=lambda v: None if v == '' else v
    )
    optional_args.add_argument(
        "--force",
        action="store_true",
        help="""
            Skip the parent top hash check and create a new revision
            even if your local state is behind the remote registry.
            """,
    )
    optional_args.add_argument(
        "--dedupe",
        action="store_true",
        help="Skip the push if the local package hash matches the remote hash.",
    )
    optional_args.add_argument(
        "--no-copy",
        action="store_true",
        help="Do not copy data. Package manifest entries will reference the data at the original location.",
    )
    push_p.set_defaults(func=cmd_push)

    return parser


def main(args=None):
    parser = create_parser()
    args = parser.parse_args(args)

    kwargs = vars(args)
    func = kwargs.pop('func')

    try:
        return func(**kwargs)
    except QuiltException as ex:
        print(ex.message, file=sys.stderr)
        return 1



import pathlib
from io import BytesIO
from unittest.mock import patch

import pandas as pd
import pytest

from quilt3 import Bucket
from quilt3.util import PhysicalKey, QuiltException

from .utils import QuiltTestCase


class TestBucket(QuiltTestCase):
    def test_bucket_construct(self):
        Bucket('s3://test-bucket')

    def test_bucket_fetch(self):
        bucket = Bucket('s3://test-bucket')

        a_contents = b'a' * 10
        b_contents = b'b' * 20

        # Fetch a directory.

        self.s3_stubber.add_response(
            method='list_objects_v2',
            service_response={
                'IsTruncated': False,
                'Contents': [
                    {'Key': 'dir/a', 'Size': len(a_contents)},
                    {'Key': 'dir/foo/b', 'Size': len(b_contents)}
                ],
            },
            expected_params={
                'Bucket': 'test-bucket',
                'Prefix': 'dir/'
            }
        )
        self.s3_stubber.add_response(
            method='get_object',
            service_response={
                'ContentLength': len(a_contents),
                'Body': BytesIO(a_contents)
            },
            expected_params={
                'Bucket': 'test-bucket',
                'Key': 'dir/a'
            }
        )
        self.s3_stubber.add_response(
            method='get_object',
            service_response={
                'ContentLength': len(b_contents),
                'Body': BytesIO(b_contents)
            },
            expected_params={
                'Bucket': 'test-bucket',
                'Key': 'dir/foo/b'
            }
        )

        with patch('quilt3.data_transfer.MAX_CONCURRENCY', 1):
            bucket.fetch('dir/', './')

        assert pathlib.Path('a').read_bytes() == a_contents
        assert pathlib.Path('foo/b').read_bytes() == b_contents

        # Fetch a single file.

        self.s3_stubber.add_response(
            method='head_object',
            service_response={
                'ContentLength': len(b_contents),
            },
            expected_params={
                'Bucket': 'test-bucket',
                'Key': 'dir/foo/b'
            }
        )

        self.s3_stubber.add_response(
            method='get_object',
            service_response={
                'ContentLength': len(b_contents),
                'Body': BytesIO(b_contents)
            },
            expected_params={
                'Bucket': 'test-bucket',
                'Key': 'dir/foo/b'
            }
        )

        bucket.fetch('dir/foo/b', './blah/')
        assert pathlib.Path('blah/b').read_bytes() == b_contents

        # Fetch a non-existent directory.

        self.s3_stubber.add_response(
            method='list_objects_v2',
            service_response={
                'IsTruncated': False
            },
            expected_params={
                'Bucket': 'test-bucket',
                'Prefix': 'does/not/exist/'
            }
        )
        with pytest.raises(QuiltException):
            bucket.fetch('does/not/exist/', './')

    def test_bucket_select(self):
        # Stubber doesn't have an accurate shape for the results of select_object_content
        chunks = [
            b'{"foo": ',
            b'9, "b',
            b'ar": 3',
            b'}\n{"foo"',
            b': 9, "bar": 1}\n{"foo": 6, "bar": 9}\n{"foo":',
            b' 1, "bar": 7}\n{"foo":',
            b' 6, "bar": 1}\n{"foo": 6, "bar": 6}',
            b'\n{"foo": 9, "bar": 6}',
            b'\n{"foo": 6, "bar": 4}\n',
            b'{"foo": 2, "bar": 0}',
            b'\n{"foo": 2, "bar": 0}\n',
            ]
        records = [{'Records': {'Payload': chunk}} for chunk in chunks]
        # noinspection PyTypeChecker
        records.append({'Stats': {
            'BytesScanned': 100,
            'BytesProcessed': 100,
            'BytesReturned': 210,
            }})
        records.append({'End': {}})

        expected_result = pd.DataFrame.from_records([
            {'foo': 9, 'bar': 3},
            {'foo': 9, 'bar': 1},
            {'foo': 6, 'bar': 9},
            {'foo': 1, 'bar': 7},
            {'foo': 6, 'bar': 1},
            {'foo': 6, 'bar': 6},
            {'foo': 9, 'bar': 6},
            {'foo': 6, 'bar': 4},
            {'foo': 2, 'bar': 0},
            {'foo': 2, 'bar': 0},
            ])

        # test normal use from extension
        expected_args = {
            'Bucket': 'test-bucket',
            'Key': 'test.json',
            'Expression': 'select * from S3Object',
            'ExpressionType': 'SQL',
            'InputSerialization': {
                'CompressionType': 'NONE',
                'JSON': {'Type': 'DOCUMENT'}
                },
            'OutputSerialization': {'JSON': {}},
            }

        boto_return_val = {'Payload': iter(records)}
        with patch.object(self.s3_client, 'select_object_content', return_value=boto_return_val) as patched:
            bucket = Bucket('s3://test-bucket')

            result = bucket.select('test.json', 'select * from S3Object')

            patched.assert_called_once_with(**expected_args)
            assert result.equals(expected_result)

    # Further testing specific to select() is in test_data_transfer

    def test_bucket_put_file(self):
        with patch("quilt3.bucket.copy_file") as copy_mock:
            bucket = Bucket('s3://test-bucket')
            bucket.put_file(key='README.md', path='./README')  # put local file to bucket

            copy_mock.assert_called_once_with(
                PhysicalKey.from_path('README'), PhysicalKey.from_url('s3://test-bucket/README.md'))

    def test_bucket_put_dir(self):
        path = pathlib.Path(__file__).parent / 'data'
        bucket = Bucket('s3://test-bucket')

        with patch("quilt3.bucket.copy_file") as copy_mock:
            bucket.put_dir('test', path)
            copy_mock.assert_called_once_with(
                PhysicalKey.from_path(str(path) + '/'), PhysicalKey.from_url('s3://test-bucket/test/'))

        with patch("quilt3.bucket.copy_file") as copy_mock:
            bucket.put_dir('test/', path)
            copy_mock.assert_called_once_with(
                PhysicalKey.from_path(str(path) + '/'), PhysicalKey.from_url('s3://test-bucket/test/'))

        with patch("quilt3.bucket.copy_file") as copy_mock:
            bucket.put_dir('', path)
            copy_mock.assert_called_once_with(
                PhysicalKey.from_path(str(path) + '/'), PhysicalKey.from_url('s3://test-bucket/'))

    def test_remote_delete(self):
        self.s3_stubber.add_response(
            method='head_object',
            service_response={},
            expected_params={
                'Bucket': 'test-bucket',
                'Key': 'file.json',
            }
        )
        self.s3_stubber.add_response(
            method='delete_object',
            service_response={},
            expected_params={
                'Bucket': 'test-bucket',
                'Key': 'file.json',
            }
        )

        bucket = Bucket('s3://test-bucket')
        bucket.delete('file.json')

        with pytest.raises(QuiltException):
            bucket.delete('dir/')

    def test_remote_delete_dir(self):
        self.s3_stubber.add_response(
            method='list_objects_v2',
            service_response={
                'IsTruncated': False,
                'Contents': [{'Key': 'dir/a'}, {'Key': 'dir/b'}],
            },
            expected_params={
                'Bucket': 'test-bucket',
                'Prefix': 'dir/'
            }
        )
        self.s3_stubber.add_response(
            method='head_object',
            service_response={},
            expected_params={
                'Bucket': 'test-bucket',
                'Key': 'dir/a'
            }
        )
        self.s3_stubber.add_response(
            method='delete_object',
            service_response={},
            expected_params={
                'Bucket': 'test-bucket',
                'Key': 'dir/a'
            }
        )
        self.s3_stubber.add_response(
            method='head_object',
            service_response={},
            expected_params={
                'Bucket': 'test-bucket',
                'Key': 'dir/b'
            }
        )
        self.s3_stubber.add_response(
            method='delete_object',
            service_response={},
            expected_params={
                'Bucket': 'test-bucket',
                'Key': 'dir/b'
            }
        )

        bucket = Bucket('s3://test-bucket')
        bucket.delete_dir('dir/')

        with pytest.raises(ValueError):
            bucket.delete_dir('dir')



"""
bucket.py

Contains the Bucket class, which provides several useful functions
    over an s3 bucket.
"""
import pathlib

from .data_transfer import (
    copy_file,
    delete_object,
    list_object_versions,
    list_objects,
    select,
)
from .search_util import search_api
from .util import PhysicalKey, QuiltException, fix_url


class Bucket:
    """Bucket interface for Quilt.
    """
    def __init__(self, bucket_uri):
        """
        Creates a Bucket object.

        Args:
            bucket_uri(str): URI of bucket to target. Must start with 's3://'

        Returns:
            A new Bucket
        """
        self._pk = PhysicalKey.from_url(bucket_uri)
        if self._pk.is_local():
            raise QuiltException("Bucket URI must be an S3 URI")
        if self._pk.path or self._pk.version_id is not None:
            raise QuiltException("Bucket URI shouldn't contain a path or a version ID")

    def search(self, query, limit=10):
        """
        Execute a search against the configured search endpoint.

        Args:
            query (str): query string to search
            limit (number): maximum number of results to return. Defaults to 10

        Query Syntax:
            By default, a normal plaintext search will be executed over the query string.
            You can use field-match syntax to filter on exact matches for fields in
                your metadata.
            The syntax for field match is `user_meta.$field_name:"exact_match"`.

        Returns:
            a list of objects with the following structure:
            ```
            [{
                "key": <key of the object>,
                "version_id": <version_id of object version>,
                "operation": <"Create" or "Delete">,
                "meta": <metadata attached to object>,
                "size": <size of object in bytes>,
                "text": <indexed text of object>,
                "source": <source document for object (what is actually stored in ElasticSeach)>,
                "time": <timestamp for operation>,
            }...]
            ```
        """
        return search_api(query, index=self._pk.bucket, limit=limit)

    def put_file(self, key, path):
        """
        Stores file at path to key in bucket.

        Args:
            key(str): key in bucket to store file at
            path(str): string representing local path to file

        Returns:
            None

        Raises:
            * if no file exists at path
            * if copy fails
        """
        dest = self._pk.join(key)
        copy_file(PhysicalKey.from_url(fix_url(path)), dest)

    def put_dir(self, key, directory):
        """
        Stores all files in the `directory` under the prefix `key`.

        Args:
            key(str): prefix to store files under in bucket
            directory(str): path to directory to grab files from

        Returns:
            None

        Raises:
            * if writing to bucket fails
        """
        # Ensure key ends in '/'.
        if key and key[-1] != '/':
            key += '/'

        src_path = pathlib.Path(directory)
        if not src_path.is_dir():
            raise QuiltException("Provided directory does not exist")

        src = PhysicalKey.from_path(str(src_path) + '/')
        dest = self._pk.join(key)
        copy_file(src, dest)

    def keys(self):
        """
        Lists all keys in the bucket.

        Returns:
            List of strings
        """
        return [x.get('Key') for x in list_objects(self._pk.bucket, '')]

    def delete(self, key):
        """
        Deletes a key from the bucket.

        Args:
            key(str): key to delete

        Returns:
            None

        Raises:
            * if delete fails
        """
        if not key:
            raise QuiltException("Must specify the key to delete")

        if key[-1] == '/':
            raise QuiltException("Must use delete_dir to delete directories")

        delete_object(self._pk.bucket, key)

    def delete_dir(self, path):
        """Delete a directory and all of its contents from the bucket.

        Parameters:
                path (str): path to the directory to delete
        """
        results = list_objects(self._pk.bucket, path)
        for result in results:
            self.delete(result['Key'])

    def ls(self, path=None, recursive=False):
        """List data from the specified path.

        Parameters:
            path (str): bucket path to list
            recursive (bool): show subdirectories and their contents as well

        Returns:
            `list`: Return value structure has not yet been permanently decided
            Currently, it's a `tuple` of `list` objects, containing the
            following: (directory info, file/object info, delete markers).
        """
        if path and not path.endswith('/'):
            path += '/'
        elif not path:
            path = ""  # enumerate top-of-bucket

        results = list_object_versions(self._pk.bucket, path, recursive=recursive)
        return results

    def fetch(self, key, path):
        """
        Fetches file (or files) at `key` to `path`.

        If `key` ends in '/', then all files with the prefix `key` will match and
        will be stored in a directory at `path`.

        Otherwise, only one file will be fetched and it will be stored at `path`.

        Args:
            key(str): key in bucket to fetch
            path(str): path in local filesystem to store file or files fetched

        Returns:
            None

        Raises:
            * if path doesn't exist
            * if download fails
        """
        source = self._pk.join(key)
        dest = PhysicalKey.from_url(fix_url(path))
        copy_file(source, dest)

    def select(self, key, query, raw=False):
        """
        Selects data from an S3 object.

        Args:
            key(str): key to query in bucket
            query(str): query to execute (SQL by default)
            query_type(str): other query type accepted by S3 service
            raw(bool): return the raw (but parsed) response

        Returns:
            pandas.DataFrame: results of query
        """
        source = self._pk.join(key)
        return select(source, query, raw=raw)



""" Testing for data_transfer.py """

import hashlib
import io
import os
import pathlib
import time
import unittest
from contextlib import redirect_stderr
from unittest import mock

import boto3
import botocore
import botocore.client
import pandas as pd
import pytest
from botocore.exceptions import ClientError, ReadTimeoutError
from botocore.stub import ANY

from quilt3 import data_transfer
from quilt3.util import PhysicalKey

from .utils import QuiltTestCase

# Code

# parquet test moved to test_formats.py

DATA_DIR = pathlib.Path(__file__).parent / 'data'


class DataTransferTest(QuiltTestCase):
    def test_select(self):
        # Note: The boto3 Stubber doesn't work properly with s3_client.select_object_content().
        #       The return value expects a dict where an iterable is in the actual results.
        chunks = [
            b'{"foo": ',
            b'9, "b',
            b'ar": 3',
            b'}\n{"foo"',
            b': 9, "bar": 1}\n{"foo": 6, "bar": 9}\n{"foo":',
            b' 1, "bar": 7}\n{"foo":',
            b' 6, "bar": 1}\n{"foo": 6, "bar": 6}',
            b'\n{"foo": 9, "bar": 6}',
            b'\n{"foo": 6, "bar": 4}\n',
            b'{"foo": 2, "bar": 0}',
            b'\n{"foo": 2, "bar": 0}\n',
            ]
        records = [{'Records': {'Payload': chunk}} for chunk in chunks]
        # noinspection PyTypeChecker
        records.append({'Stats': {
            'BytesScanned': 100,
            'BytesProcessed': 100,
            'BytesReturned': 210,
            }})
        records.append({'End': {}})

        expected_result = pd.DataFrame.from_records([
            {'foo': 9, 'bar': 3},
            {'foo': 9, 'bar': 1},
            {'foo': 6, 'bar': 9},
            {'foo': 1, 'bar': 7},
            {'foo': 6, 'bar': 1},
            {'foo': 6, 'bar': 6},
            {'foo': 9, 'bar': 6},
            {'foo': 6, 'bar': 4},
            {'foo': 2, 'bar': 0},
            {'foo': 2, 'bar': 0},
            ])

        # test normal use from extension
        expected_args = {
            'Bucket': 'foo',
            'Key': 'bar/baz.json',
            'Expression': 'select * from S3Object',
            'ExpressionType': 'SQL',
            'InputSerialization': {
                'CompressionType': 'NONE',
                'JSON': {'Type': 'DOCUMENT'}
                },
            'OutputSerialization': {'JSON': {}},
            }
        boto_return_val = {'Payload': iter(records)}
        with mock.patch.object(self.s3_client, 'select_object_content', return_value=boto_return_val) as patched:
            result = data_transfer.select(PhysicalKey.from_url('s3://foo/bar/baz.json'), 'select * from S3Object')

            patched.assert_called_once_with(**expected_args)
            assert result.equals(expected_result)

        with mock.patch.object(self.s3_client, 'select_object_content'):
            # No format determined.
            with pytest.raises(data_transfer.QuiltException):
                result = data_transfer.select(PhysicalKey.from_url('s3://foo/bar/baz'), 'select * from S3Object')

        # test format-specified in metadata
        expected_args = {
            'Bucket': 'foo',
            'Key': 'bar/baz',
            'Expression': 'select * from S3Object',
            'ExpressionType': 'SQL',
            'InputSerialization': {
                'CompressionType': 'NONE',
                'JSON': {'Type': 'DOCUMENT'}
            },
            'OutputSerialization': {'JSON': {}},
        }

        boto_return_val = {'Payload': iter(records)}
        with mock.patch.object(self.s3_client, 'select_object_content', return_value=boto_return_val) as patched:
            result = data_transfer.select(PhysicalKey.from_url('s3://foo/bar/baz'), 'select * from S3Object',
                                          meta={'target': 'json'})
            assert result.equals(expected_result)
            patched.assert_called_once_with(**expected_args)

        # test compression is specified
        expected_args = {
            'Bucket': 'foo',
            'Key': 'bar/baz.json.gz',
            'Expression': 'select * from S3Object',
            'ExpressionType': 'SQL',
            'InputSerialization': {
                'CompressionType': 'GZIP',
                'JSON': {'Type': 'DOCUMENT'}
                },
            'OutputSerialization': {'JSON': {}},
            }
        boto_return_val = {'Payload': iter(records)}
        with mock.patch.object(self.s3_client, 'select_object_content', return_value=boto_return_val) as patched:
            # result ignored -- returned data isn't compressed, and this has already been tested.
            data_transfer.select(PhysicalKey.from_url('s3://foo/bar/baz.json.gz'), 'select * from S3Object')
            patched.assert_called_once_with(**expected_args)

    def test_get_size_and_version(self):
        response = {
            'ETag': '12345',
            'VersionId': '1.0',
            'ContentLength': 123,
        }
        expected_params = {
            'Bucket': 'my_bucket',
            'Key': 'my_obj',
        }
        self.s3_stubber.add_response('head_object', response, expected_params)

        # Verify the verion is present
        assert data_transfer.get_size_and_version(PhysicalKey.from_url('s3://my_bucket/my_obj'))[1] == '1.0'

    def test_list_local_url(self):
        dir_path = DATA_DIR / 'dir'
        contents = set(list(data_transfer.list_url(PhysicalKey.from_path(dir_path))))
        assert contents == set([
            ('foo.txt', 4),
            ('x/blah.txt', 6)
        ])

    def test_etag(self):
        assert data_transfer._calculate_etag(DATA_DIR / 'small_file.csv') == '"0bec5bf6f93c547bc9c6774acaf85e1a"'
        assert data_transfer._calculate_etag(
            DATA_DIR / 'buggy_parquet.parquet') == '"dfb5aca048931d396f4534395617363f"'

    def test_simple_upload(self):
        path = DATA_DIR / 'small_file.csv'

        # Unversioned bucket
        self.s3_stubber.add_response(
            method='put_object',
            service_response={
            },
            expected_params={
                'Body': ANY,
                'Bucket': 'example',
                'Key': 'foo.csv',
            }
        )

        data_transfer.copy_file(PhysicalKey.from_path(path), PhysicalKey.from_url('s3://example/foo.csv'))

    def test_multi_upload(self):
        path1 = DATA_DIR / 'small_file.csv'
        path2 = DATA_DIR / 'dir/foo.txt'

        # Unversioned bucket
        self.s3_stubber.add_response(
            method='put_object',
            service_response={
            },
            expected_params={
                'Body': ANY,
                'Bucket': 'example1',
                'Key': 'foo.csv',
            }
        )

        # Versioned bucket
        self.s3_stubber.add_response(
            method='put_object',
            service_response={
                'VersionId': 'v123'
            },
            expected_params={
                'Body': ANY,
                'Bucket': 'example2',
                'Key': 'foo.txt',
            }
        )

        # stubber expects responses in order, so disable multi-threading.
        with mock.patch('quilt3.data_transfer.MAX_CONCURRENCY', 1):
            urls = data_transfer.copy_file_list([
                (PhysicalKey.from_path(path1), PhysicalKey.from_url('s3://example1/foo.csv'), path1.stat().st_size),
                (PhysicalKey.from_path(path2), PhysicalKey.from_url('s3://example2/foo.txt'), path2.stat().st_size),
            ])

            assert urls[0] == PhysicalKey.from_url('s3://example1/foo.csv')
            assert urls[1] == PhysicalKey.from_url('s3://example2/foo.txt?versionId=v123')

    def test_upload_large_file(self):
        path = DATA_DIR / 'large_file.npy'

        self.s3_stubber.add_client_error(
            method='head_object',
            http_status_code=404,
            expected_params={
                'Bucket': 'example',
                'Key': 'large_file.npy',
            }
        )

        self.s3_stubber.add_response(
            method='put_object',
            service_response={
                'VersionId': 'v1'
            },
            expected_params={
                'Body': ANY,
                'Bucket': 'example',
                'Key': 'large_file.npy',
            }
        )

        urls = data_transfer.copy_file_list([
            (PhysicalKey.from_path(path), PhysicalKey.from_url('s3://example/large_file.npy'), path.stat().st_size),
        ])
        assert urls[0] == PhysicalKey.from_url('s3://example/large_file.npy?versionId=v1')

    def test_upload_large_file_etag_match(self):
        path = DATA_DIR / 'large_file.npy'

        self.s3_stubber.add_response(
            method='head_object',
            service_response={
                'ContentLength': path.stat().st_size,
                'ETag': data_transfer._calculate_etag(path),
                'VersionId': 'v1',
            },
            expected_params={
                'Bucket': 'example',
                'Key': 'large_file.npy',
            }
        )

        urls = data_transfer.copy_file_list([
            (PhysicalKey.from_path(path), PhysicalKey.from_url('s3://example/large_file.npy'), path.stat().st_size),
        ])
        assert urls[0] == PhysicalKey.from_url('s3://example/large_file.npy?versionId=v1')

    def test_upload_large_file_etag_mismatch(self):
        path = DATA_DIR / 'large_file.npy'

        self.s3_stubber.add_response(
            method='head_object',
            service_response={
                'ContentLength': path.stat().st_size,
                'ETag': '"123"',
                'VersionId': 'v1',
            },
            expected_params={
                'Bucket': 'example',
                'Key': 'large_file.npy',
            }
        )

        self.s3_stubber.add_response(
            method='put_object',
            service_response={
                'VersionId': 'v2'
            },
            expected_params={
                'Body': ANY,
                'Bucket': 'example',
                'Key': 'large_file.npy',
            }
        )

        urls = data_transfer.copy_file_list([
            (PhysicalKey.from_path(path), PhysicalKey.from_url('s3://example/large_file.npy'), path.stat().st_size),
        ])
        assert urls[0] == PhysicalKey.from_url('s3://example/large_file.npy?versionId=v2')

    def test_multipart_upload(self):
        name = 'very_large_file.bin'
        path = pathlib.Path(name)

        size = 30 * 1024 * 1024
        chunksize = 8 * 1024 * 1024

        chunks = -(-size // chunksize)

        # Create an empty 30MB file; shouldn't take up any actual space on any reasonable filesystem.
        with open(path, 'wb') as fd:
            fd.seek(size - 1)
            fd.write(b'!')

        self.s3_stubber.add_client_error(
            method='head_object',
            http_status_code=404,
            expected_params={
                'Bucket': 'example',
                'Key': name,
            }
        )

        self.s3_stubber.add_response(
            method='create_multipart_upload',
            service_response={
                'UploadId': '123'
            },
            expected_params={
                'Bucket': 'example',
                'Key': name,
            }
        )

        for part_num in range(1, chunks+1):
            self.s3_stubber.add_response(
                method='upload_part',
                service_response={
                    'ETag': 'etag%d' % part_num
                },
                expected_params={
                    'Bucket': 'example',
                    'Key': name,
                    'UploadId': '123',
                    'Body': ANY,
                    'PartNumber': part_num
                }
            )

        self.s3_stubber.add_response(
            method='complete_multipart_upload',
            service_response={},
            expected_params={
                'Bucket': 'example',
                'Key': name,
                'UploadId': '123',
                'MultipartUpload': {
                    'Parts': [{
                        'ETag': 'etag%d' % i,
                        'PartNumber': i
                    } for i in range(1, chunks+1)]
                }
            }
        )

        with mock.patch('quilt3.data_transfer.MAX_CONCURRENCY', 1):
            data_transfer.copy_file_list([
                (PhysicalKey.from_path(path), PhysicalKey.from_url(f's3://example/{name}'), path.stat().st_size),
            ])

    def test_multipart_copy(self):
        size = 100 * 1024 * 1024 * 1024

        # size / 8MB would give us 12501 chunks - but the maximum allowed is 10000,
        # so we should end with 16MB chunks instead.
        chunksize = 8 * 1024 * 1024
        assert size / chunksize > 10000
        chunksize *= 2

        chunks = -(-size // chunksize)
        assert chunks <= 10000

        self.s3_stubber.add_response(
            method='create_multipart_upload',
            service_response={
                'UploadId': '123'
            },
            expected_params={
                'Bucket': 'example2',
                'Key': 'large_file2.npy',
            }
        )

        for part_num in range(1, chunks+1):
            self.s3_stubber.add_response(
                method='upload_part_copy',
                service_response={
                    'CopyPartResult': {
                        'ETag': 'etag%d' % part_num
                    }
                },
                expected_params={
                    'Bucket': 'example2',
                    'Key': 'large_file2.npy',
                    'UploadId': '123',
                    'PartNumber': part_num,
                    'CopySource': {
                        'Bucket': 'example1',
                        'Key': 'large_file1.npy'
                    },
                    'CopySourceRange': 'bytes=%d-%d' % (
                        (part_num-1) * chunksize,
                        min(part_num * chunksize, size) - 1
                    )
                }
            )

        self.s3_stubber.add_response(
            method='complete_multipart_upload',
            service_response={},
            expected_params={
                'Bucket': 'example2',
                'Key': 'large_file2.npy',
                'UploadId': '123',
                'MultipartUpload': {
                    'Parts': [{
                        'ETag': 'etag%d' % i,
                        'PartNumber': i
                    } for i in range(1, chunks+1)]
                }
            }
        )

        with mock.patch('quilt3.data_transfer.MAX_CONCURRENCY', 1):
            stderr = io.StringIO()

            with redirect_stderr(stderr), mock.patch('quilt3.data_transfer.DISABLE_TQDM', False):
                data_transfer.copy_file_list([
                    (
                        PhysicalKey.from_url('s3://example1/large_file1.npy'),
                        PhysicalKey.from_url('s3://example2/large_file2.npy'),
                        size
                    ),
                ])
            assert stderr.getvalue()

    @mock.patch('botocore.client.BaseClient._make_api_call')
    def test_calculate_sha256_read_timeout(self, mocked_api_call):
        bucket = 'test-bucket'
        key = 'dir/a'
        vid = 'a1234'

        a_contents = b'a' * 10

        pk = PhysicalKey(bucket, key, vid)
        exc = ReadTimeoutError('Error Uploading', endpoint_url="s3://foobar")
        mocked_api_call.side_effect = exc
        results = data_transfer.calculate_sha256([pk], [len(a_contents)])
        assert mocked_api_call.call_count == data_transfer.MAX_FIX_HASH_RETRIES
        assert results == [exc]

    def test_copy_file_list_retry(self):
        bucket = 'test-bucket'
        other_bucket = f'{bucket}-other'
        key = 'dir/a'
        vid = None

        src = PhysicalKey(bucket, key, vid)
        dst = PhysicalKey(other_bucket, key, vid)

        with mock.patch('botocore.client.BaseClient._make_api_call',
                        side_effect=ClientError({}, 'CopyObject')) as mocked_api_call:
            with pytest.raises(ClientError):
                data_transfer.copy_file_list([(src, dst, 1)])
            self.assertEqual(mocked_api_call.call_count, data_transfer.MAX_COPY_FILE_LIST_RETRIES)

    def test_copy_file_list_retry_non_client_error(self):
        """
        copy_file_list() is not retrying on random exceptions.
        """
        bucket = 'test-bucket'
        other_bucket = f'{bucket}-other'
        key = 'dir/a'
        vid = None

        src = PhysicalKey(bucket, key, vid)
        dst = PhysicalKey(other_bucket, key, vid)

        with mock.patch('botocore.client.BaseClient._make_api_call',
                        side_effect=Exception('test exception')) as mocked_api_call:
            with pytest.raises(Exception, match='test exception'):
                data_transfer.copy_file_list([(src, dst, 1)])
            assert mocked_api_call.call_count == 1

    def test_copy_file_list_multipart_retry(self):
        bucket = 'test-bucket'
        other_bucket = f'{bucket}-other'
        key = 'dir/a'
        vid = None

        src = PhysicalKey(bucket, key, vid)
        dst = PhysicalKey(other_bucket, key, vid)
        parts = 2 * data_transfer.s3_transfer_config.max_request_concurrency
        size = parts * data_transfer.s3_transfer_config.multipart_threshold

        def side_effect(operation_name, *args, **kwargs):
            if operation_name == 'CreateMultipartUpload':
                return {'UploadId': '123'}
            time.sleep(0.1)
            raise ClientError({}, 'CopyObject')

        with mock.patch('botocore.client.BaseClient._make_api_call', side_effect=side_effect):
            with pytest.raises(ClientError):
                data_transfer.copy_file_list([(src, dst, size)])

    @mock.patch.multiple(
        'quilt3.data_transfer.s3_transfer_config',
        multipart_threshold=1,
        multipart_chunksize=1,
    )
    @mock.patch('quilt3.data_transfer.MAX_CONCURRENCY', 1)
    def test_download_latest_in_versioned_bucket(self):
        bucket = 'example'
        key = 'foo.csv'
        src = PhysicalKey(bucket, key, None)
        latest_version = '1'
        latest_size = 3

        # Check what is the latest version and size.
        expected_params = {
            'Bucket': bucket,
            'Key': key,
        }
        self.s3_stubber.add_response(
            'head_object',
            service_response={
                'VersionId': latest_version,
                'ContentLength': latest_size,
            },
            expected_params=expected_params,
        )
        for i in range(latest_size):
            self.s3_stubber.add_response(
                'get_object',
                service_response={
                    'Body': io.BytesIO(b'0'),
                },
                expected_params={
                    **expected_params,
                    'Range': f'bytes={i}-{i}',
                    # Version must be specified, otherwise we will end with
                    # a truncated file if the file was modified after getting the latest
                    # version/size.
                    'VersionId': latest_version,
                },
            )

        data_transfer.copy_file(
            src,
            PhysicalKey.from_path('some-file'),
        )


class Success(Exception):
    pass


class S3UploadProgressTest(unittest.TestCase):
    def setUp(self):
        self.s3_client = boto3.client('s3', config=botocore.client.Config(signature_version=botocore.UNSIGNED))
        self.s3_client_patcher = mock.patch.multiple(
            'quilt3.data_transfer.S3ClientProvider',
            _build_client=lambda *args, **kwargs: self.s3_client,
            client_type_known=lambda *args, **kwargs: True,
        )
        self.s3_client_patcher.start()
        self.addCleanup(self.s3_client_patcher.stop)

    def test_body_is_seekable(self):
        """
        No errors if request body.read() or body.seek() are called right before sending request.
        """
        def handler(request, **kwargs):
            request.body.read(2)
            request.body.seek(0)

            raise Success

        path = DATA_DIR / 'small_file.csv'
        self.s3_client.meta.events.register_first('before-send.*', handler)
        with pytest.raises(Success):
            data_transfer.copy_file(PhysicalKey.from_path(path), PhysicalKey.from_url('s3://example/foo.csv'))

    @mock.patch('tqdm.tqdm.update')
    def test_progress_updateds(self, mocked_update):
        """
        Progress callback is called when calling body.read() or body.seek().
        """

        def handler(request, **kwargs):
            request.body.read(2)
            mocked_update.assert_called_once_with(2)

            mocked_update.reset_mock()
            request.body.seek(0)
            mocked_update.assert_called_once_with(-2)

            raise Success

        path = DATA_DIR / 'small_file.csv'
        self.s3_client.meta.events.register_first('before-send.*', handler)
        with pytest.raises(Success):
            data_transfer.copy_file(PhysicalKey.from_path(path), PhysicalKey.from_url('s3://example/foo.csv'))


class S3DownloadTest(QuiltTestCase):
    data = b'0123456789abcdef'
    size = len(data)

    bucket = 'test-bucket'
    key = 'test-key'
    src = PhysicalKey(bucket, key, None)

    filename = 'some-file-name'
    dst = PhysicalKey(None, filename, None)

    def _test_download(self, *, threshold, chunksize, parts=data, devnull=False):
        dst = PhysicalKey(None, os.devnull, None) if devnull else self.dst

        with self.s3_test_multi_thread_download(
            self.bucket, self.key, parts, threshold=threshold, chunksize=chunksize
        ):
            data_transfer.copy_file_list([(self.src, dst, self.size)])

        if not devnull:
            with open(self.filename, 'rb') as f:
                assert f.read() == self.data

    def test_threshold_gt_size(self):
        self._test_download(threshold=self.size + 1, chunksize=5)

    def test_threshold_eq_size(self):
        parts = {
            'bytes=0-4':  self.data[:5],
            'bytes=5-9': self.data[5:10],
            'bytes=10-14': self.data[10:15],
            'bytes=15-15': self.data[15:],
        }
        self._test_download(threshold=self.size, chunksize=5, parts=parts)

    def test_threshold_eq_size_special_file(self):
        if os.name == 'nt':
            with pytest.raises(ValueError, match=f'Cannot download to {os.devnull!r}: reserved file name'):
                self._test_download(threshold=self.size, chunksize=5, devnull=True)
        else:
            self._test_download(threshold=self.size, chunksize=5, devnull=True)

    def test_threshold_eq_chunk_eq_size(self):
        self._test_download(threshold=self.size, chunksize=self.size)

    def test_threshold_eq_chunk_gt_size(self):
        self._test_download(threshold=self.size, chunksize=self.size + 1)


class S3HashingTest(QuiltTestCase):
    data = b'0123456789abcdef'
    size = len(data)
    hasher = hashlib.sha256

    bucket = 'test-bucket'
    key = 'test-key'
    src = PhysicalKey(bucket, key, None)

    def _hashing_subtest(self, *, threshold, chunksize, data=data):
        with self.s3_test_multi_thread_download(
            self.bucket, self.key, data, threshold=threshold, chunksize=chunksize
        ):
            assert data_transfer.calculate_sha256([self.src], [self.size]) == [self.hasher(self.data).hexdigest()]

    def test_single_request(self):
        params = (
            (self.size + 1, 5),
            (self.size, self.size),
            (self.size, self.size + 1),
            (5, self.size),
        )
        for threshold, chunksize in params:
            with self.subTest(threshold=threshold, chunksize=chunksize):
                self._hashing_subtest(threshold=threshold, chunksize=chunksize)

    def test_multi_request(self):
        params = (
            (
                self.size, 5, {
                    'bytes=0-4': self.data[:5],
                    'bytes=5-9': self.data[5:10],
                    'bytes=10-14': self.data[10:15],
                    'bytes=15-15': self.data[15:],
                }
            ),
            (
                5, self.size - 1, {
                    'bytes=0-14': self.data[:15],
                    'bytes=15-15': self.data[15:],
                }
            ),
        )
        for threshold, chunksize, data in params:
            for concurrency in (len(data), 1):
                with mock.patch('quilt3.data_transfer.s3_transfer_config.max_request_concurrency', concurrency):
                    with self.subTest(threshold=threshold, chunksize=chunksize, data=data, concurrency=concurrency):
                        self._hashing_subtest(threshold=threshold, chunksize=chunksize, data=data)



import abc
import operator
import time

from quilt3.data_transfer import (
    copy_file,
    delete_url,
    get_bytes,
    list_url,
    put_bytes,
)
from quilt3.util import PhysicalKey, QuiltException


class PackageRegistry(abc.ABC):
    latest_tag_name = 'latest'
    top_hash_len = 64
    revision_pointers = False
    workflow_conf_path = '.quilt/workflows/config.yml'

    def __init__(self, base: PhysicalKey):
        self.base = base
        self.root = self.base.join(self.root_path)

    def __repr__(self):
        return f'<{self.__class__.__name__}({self.base})>'

    def __eq__(self, other):
        return other.__class__ == self.__class__ and other.base == self.base

    @property
    def is_local(self) -> bool:
        return self.base.is_local()

    @property
    @abc.abstractmethod
    def root_path(self):
        pass

    @property
    @abc.abstractmethod
    def pointers_global_dir(self) -> PhysicalKey:
        pass

    @abc.abstractmethod
    def pointers_dir(self, pkg_name: str) -> PhysicalKey:
        pass

    @abc.abstractmethod
    def pointer_pk(self, pkg_name: str, pointer_name: str) -> PhysicalKey:
        pass

    def pointer_latest_pk(self, pkg_name: str) -> PhysicalKey:
        return self.pointer_pk(pkg_name, self.latest_tag_name)

    @property
    @abc.abstractmethod
    def manifests_global_dir(self) -> PhysicalKey:
        pass

    @abc.abstractmethod
    def manifests_package_dir(self, pkg_name: str) -> PhysicalKey:
        pass

    @abc.abstractmethod
    def manifest_pk(self, pkg_name: str, top_hash: str) -> PhysicalKey:
        pass

    @abc.abstractmethod
    def list_packages(self):
        pass

    @abc.abstractmethod
    def list_package_pointers(self, pkg_name: str):
        pass

    @abc.abstractmethod
    def list_package_versions(self, pkg_name: str):
        pass

    @abc.abstractmethod
    def delete_package(self, pkg_name: str):
        pass

    @abc.abstractmethod
    def delete_package_version(self, pkg_name: str, top_hash: str):
        pass

    @abc.abstractmethod
    def push_manifest(self, pkg_name: str, top_hash: str, manifest_data: bytes):
        pass

    @abc.abstractmethod
    def resolve_top_hash(self, pkg_name: str, hash_prefix: str) -> str:
        pass

    @abc.abstractmethod
    def shorten_top_hash(self, pkg_name: str, top_hash: str) -> str:
        pass

    @property
    def workflow_conf_pk(self) -> PhysicalKey:
        return self.base.join(self.workflow_conf_path)

    def get_workflow_config(self):
        from quilt3.workflows import WorkflowConfig
        return WorkflowConfig.load(self.workflow_conf_pk)


class PackageRegistryV1(PackageRegistry):
    root_path = '.quilt'
    revision_pointers = True

    @property
    def pointers_global_dir(self) -> PhysicalKey:
        return self.root.join('named_packages/')

    def pointers_dir(self, pkg_name: str) -> PhysicalKey:
        return self.root.join(f'named_packages/{pkg_name}/')

    def _pointers_usr_dir(self, pkg_name: str) -> PhysicalKey:
        return self.root.join(f'named_packages/{pkg_name.partition("/")[0]}/')

    def pointer_pk(self, pkg_name: str, pointer_name: str) -> PhysicalKey:
        return self.root.join(f'named_packages/{pkg_name}/{pointer_name}')

    @property
    def manifests_global_dir(self) -> PhysicalKey:
        return self.root.join('packages/')

    def manifests_package_dir(self, pkg_name: str) -> PhysicalKey:
        return self.manifests_global_dir  # TODO: does this make sense?

    def manifest_pk(self, pkg_name: str, top_hash: str) -> PhysicalKey:
        return self.root.join(f'packages/{top_hash}')

    def push_manifest(self, pkg_name: str, top_hash: str, manifest_data: bytes):
        """returns: timestamp to support catalog drag-and-drop => browse"""
        put_bytes(manifest_data, self.manifest_pk(pkg_name, top_hash))
        hash_bytes = top_hash.encode()
        # TODO: use a float to string formatter instead of double casting
        timestamp_str = str(int(time.time()))
        put_bytes(hash_bytes, self.pointer_pk(pkg_name, timestamp_str))
        put_bytes(hash_bytes, self.pointer_latest_pk(pkg_name))
        return timestamp_str

    @staticmethod
    def _top_hash_from_path(path: str) -> str:
        return path

    def resolve_top_hash(self, pkg_name: str, hash_prefix: str) -> str:
        if len(hash_prefix) == 64:
            top_hash = hash_prefix
        elif 6 <= len(hash_prefix) < 64:
            matching_hashes = [self._top_hash_from_path(h) for h, _
                               in list_url(self.manifests_package_dir(pkg_name))
                               if h.startswith(hash_prefix)]
            if not matching_hashes:
                raise QuiltException("Found zero matches for %r" % hash_prefix)
            elif len(matching_hashes) > 1:
                raise QuiltException("Found multiple matches: %r" % hash_prefix)
            else:
                top_hash = matching_hashes[0]
        else:
            raise QuiltException("Invalid hash: %r" % hash_prefix)
        # TODO: verify that name is correct with respect to this top_hash
        return top_hash

    resolve_top_hash_requires_pkg_name = False

    def shorten_top_hash(self, pkg_name: str, top_hash: str) -> str:
        min_shorthash_len = 7

        matches = [self._top_hash_from_path(h) for h, _ in list_url(self.manifests_package_dir(pkg_name))
                   if h.startswith(top_hash[:min_shorthash_len])]
        if len(matches) == 0:
            raise ValueError(f"Tophash {top_hash} was not found in registry {self.base}")
        for prefix_length in range(min_shorthash_len, 64):
            potential_shorthash = top_hash[:prefix_length]
            matches = [h for h in matches if h.startswith(potential_shorthash)]
            if len(matches) == 1:
                return potential_shorthash

    def list_package_versions(self, pkg_name: str):
        return self.list_package_pointers(pkg_name)

    def delete_package_version(self, pkg_name: str, top_hash: str):
        deleted = []
        remaining = []
        for path, pkg_hash in self.list_package_versions(pkg_name):
            (deleted if pkg_hash == top_hash else remaining).append(path)
        if not deleted:
            raise QuiltException("No such package version exists in the given directory.")
        for path in deleted:
            delete_url(self.pointer_pk(pkg_name, path))
        if 'latest' in deleted and remaining:
            # Create a new "latest". Technically, we need to compare numerically,
            # but string comparisons will be fine till year 2286.
            new_latest = max(remaining)
            copy_file(self.pointer_pk(pkg_name, new_latest), self.pointer_latest_pk(pkg_name))


class PackageRegistryV2(PackageRegistry):
    """
    All metadata files live under `.quilt/v2`:
    * the manifests use `.quilt/v2/manifests/<usr>@<pkg>/<top_hash>/manifest.jsonl` format for path
    * the tag files (or pointers) live under `.quilt/v2/tags/<usr>@<pkg>/<tag_name>`, each of these files
      contains the top hash of one of package manifests.
    """

    root_path = '.quilt/v2'

    _package_name_to_path = operator.methodcaller('replace', '/', '@')

    @property
    def pointers_global_dir(self) -> PhysicalKey:
        return self.root.join('tags/')

    def pointers_dir(self, pkg_name: str) -> PhysicalKey:
        return self.root.join(f'tags/{self._package_name_to_path(pkg_name)}/')

    def pointer_pk(self, pkg_name: str, pointer_name: str) -> PhysicalKey:
        return self.root.join(f'tags/{self._package_name_to_path(pkg_name)}/{pointer_name}')

    @property
    def manifests_global_dir(self) -> PhysicalKey:
        return self.root.join('manifests/')

    def manifests_package_dir(self, pkg_name: str) -> PhysicalKey:
        return self.root.join(f'manifests/{self._package_name_to_path(pkg_name)}/')

    def manifest_pk(self, pkg_name: str, top_hash: str) -> PhysicalKey:
        return self.root.join(f'manifests/{self._package_name_to_path(pkg_name)}/{top_hash}/manifest.jsonl')

    def _manifest_parent_pk(self, pkg_name: str, top_hash: str) -> PhysicalKey:
        return self.root.join(f'manifests/{self._package_name_to_path(pkg_name)}/{top_hash}/')

    @abc.abstractmethod
    def list_package_versions_with_timestamps(self, pkg_name: str):
        """
        Returns:
            An iterable of tuples containing the datetime and hash for the package.
        """

    def list_package_versions(self, pkg_name: str):
        for dt, top_hash in self.list_package_versions_with_timestamps(pkg_name):
            yield str(int(dt.timestamp())), top_hash

    def push_manifest(self, pkg_name: str, top_hash: str, manifest_data: bytes):
        put_bytes(manifest_data, self.manifest_pk(pkg_name, top_hash))
        put_bytes(top_hash.encode(), self.pointer_latest_pk(pkg_name))

    @staticmethod
    def _top_hash_from_path(path: str) -> str:
        return path.rsplit('/', 2)[-2]

    resolve_top_hash = PackageRegistryV1.resolve_top_hash
    resolve_top_hash_requires_pkg_name = True
    shorten_top_hash = PackageRegistryV1.shorten_top_hash

    def delete_package_version(self, pkg_name: str, top_hash: str):
        delete_url(self.manifest_pk(pkg_name, top_hash))
        if get_bytes(self.pointer_latest_pk(pkg_name)).decode() == top_hash:
            delete_url(self.pointer_latest_pk(pkg_name))
            timestamp, new_latest = max(self.list_package_versions_with_timestamps(pkg_name), default=(None, None))
            if new_latest:
                put_bytes(new_latest.encode(), self.pointer_latest_pk(pkg_name))
        for pointer, pointer_top_hash in self.list_package_pointers(pkg_name):
            if pointer_top_hash == top_hash:
                delete_url(self.pointer_pk(pkg_name, pointer))



""" Testing for util.py """

import pathlib
from unittest import mock

import pytest

from quilt3 import util

# Constants
TEST_YAML = """
    # This is an arbitrary comment solely for the purposes of testing.
    c: the speed of light
    d: a programming language that almost mattered
    # Another arbitrary comment.
    e: a not-so-hip MC from a relatively unknown nightclub    # do you like cats?
    """


# Code
def test_write_yaml(tmpdir):
    fname = tmpdir / 'some_file.yml'

    util.write_yaml({'testing': 'foo'}, fname)
    assert fname.read_text('utf-8') == 'testing: foo\n'

    util.write_yaml({'testing': 'bar'}, fname, keep_backup=True)

    fnames = [name for name in tmpdir.listdir() if name.basename.startswith('some_file')]

    assert len(fnames) == 2

    fname2 = max(fnames, key=lambda x: len(x.basename))  # backup name is longer
    assert fname2.read_text('utf-8') == 'testing: foo\n'
    assert fname.read_text('utf-8') == 'testing: bar\n'


def test_read_yaml(tmpdir):
    # Read a string
    parsed_string = util.read_yaml(TEST_YAML)
    fname = tmpdir / 'test_read_yaml.yml'

    util.write_yaml(parsed_string, fname)

    # Read file descriptor..
    with fname.open('r') as f:
        parsed_file = util.read_yaml(f)
    assert parsed_file == parsed_string

    # Read Path object
    parsed_path_obj = util.read_yaml(pathlib.Path(fname))
    assert parsed_string == parsed_path_obj


def test_read_yaml_exec_flaw(capfd):
    # We don't execute anything remote, but someone could give a bad build.yml..
    with pytest.raises(util.QuiltException) as exc_info:
        util.read_yaml("""!!python/object/apply:os.system\nargs: ['echo arbitrary code execution!']""")
    assert "could not determine a constructor for the tag" in str(exc_info.value)


def test_validate_url():
    with pytest.raises(util.QuiltException, match='Port must be a number'):
        util.validate_url('http://foo:bar')

    with pytest.raises(util.QuiltException, match='Requires at least scheme and host'):
        util.validate_url('blah')


@pytest.mark.parametrize(
    'env_val, expected_val',
    (
        (None, None),
        ('', None),
        ('1', 1),
        ('10', 10),
        ('20', 20),
    )
)
def test_get_pos_int_from_env(env_val, expected_val):
    var_name = 'ENV_VAR_NAME'
    with mock.patch.dict('os.environ', {} if env_val is None else {var_name: env_val}, clear=True):
        assert util.get_pos_int_from_env(var_name) == expected_val


@pytest.mark.parametrize('env_val', ('blah', '-3', '0'))
def test_get_pos_int_from_env_error(env_val):
    var_name = 'ENV_VAR_NAME'
    with mock.patch.dict('os.environ', {} if env_val is None else {var_name: env_val}, clear=True):
        with pytest.raises(ValueError) as e:
            util.get_pos_int_from_env(var_name)
        assert f'{var_name} must be a positive integer' == str(e.value)



from unittest import mock
from unittest.mock import patch

import pytest

from quilt3 import main

from .utils import QuiltTestCase

create_parser = main.create_parser


patch_package_class = mock.patch('quilt3.main.Package', autospec=True)


class CommandLineTestCase(QuiltTestCase):
    """
    Base TestCase class, sets up a CLI parser
    """
    @classmethod
    def setUpClass(cls):
        parser = create_parser()
        cls.parser = parser


class QuiltCLITestCase(CommandLineTestCase):
    def test_quilt_config(self):
        args = self.parser.parse_args(['config', 'https://foo.bar'])
        assert args.catalog_url == 'https://foo.bar'

    def test_push(self):
        name = 'test/name'
        dir_path = 'test/dir/path'

        with patch_package_class as mocked_package_class:
            mocked_package_class.browse.side_effect = FileNotFoundError()

            main.main(('push', '--dir', dir_path, name))

            mocked_package_class.browse.assert_called_once_with(name, None)
            mocked_package_class.assert_called_once_with()
            mocked_package = mocked_package_class.return_value
            mocked_package.set_dir.assert_called_once_with('.', dir_path, meta=None)
            mocked_package.push.assert_called_once_with(
                name, registry=None, dest=None, message=None, workflow=..., force=False, dedupe=False
            )

    def test_push_force(self):
        name = 'test/name'
        dir_path = 'test/dir/path'

        with patch_package_class as mocked_package_class:
            mocked_package_class.browse.side_effect = FileNotFoundError()

            main.main(('push', '--dir', dir_path, name, '--force'))

            mocked_package_class.browse.assert_called_once_with(name, None)
            mocked_package_class.assert_called_once_with()
            mocked_package = mocked_package_class.return_value
            mocked_package.set_dir.assert_called_once_with('.', dir_path, meta=None)
            mocked_package.push.assert_called_once_with(
                name, registry=None, dest=None, message=None, workflow=..., force=True, dedupe=False
            )

    def test_push_dedupe(self):
        name = 'test/name'
        dir_path = 'test/dir/path'

        with patch_package_class as mocked_package_class:
            mocked_package_class.browse.side_effect = FileNotFoundError()

            main.main(('push', '--dir', dir_path, name, '--dedupe'))

            mocked_package_class.browse.assert_called_once_with(name, None)
            mocked_package_class.assert_called_once_with()
            mocked_package = mocked_package_class.return_value
            mocked_package.set_dir.assert_called_once_with('.', dir_path, meta=None)
            mocked_package.push.assert_called_once_with(
                name, registry=None, dest=None, message=None, workflow=..., force=False, dedupe=True
            )

    def test_push_existing(self):
        name = 'test/name'
        dir_path = 'test/dir/path'

        with patch_package_class as mocked_package_class:
            main.main(('push', '--dir', dir_path, name))

            mocked_package_class.browse.assert_called_once_with(name, None)
            mocked_package_class.assert_not_called()
            mocked_package = mocked_package_class.browse.return_value
            mocked_package.set_dir.assert_called_once_with('.', dir_path, meta=None)
            mocked_package.push.assert_called_once_with(
                name, registry=None, dest=None, message=None, workflow=..., force=False, dedupe=False
            )


@pytest.mark.parametrize(
    'meta_data, expected_meta',
    [
        (None, None),
        ('{"test": "meta"}', {"test": "meta"}),
    ]
)
def test_push_with_meta_data(
    meta_data,
    expected_meta,
):
    name = 'test/name'
    dir_path = 'test/dir/path'

    with patch_package_class as mocked_package_class, \
         mock.patch('quilt3.main.parse_arg_json', wraps=main.parse_arg_json) as mocked_parse_json_arg:
        mocked_package_class.browse.side_effect = FileNotFoundError()

        # '--registry' defaults to configured remote registry hence optional.
        if meta_data:
            main.main(('push', '--dir', dir_path, name, '--meta', meta_data))
            mocked_parse_json_arg.assert_called_once_with(meta_data)
        else:
            main.main(('push', '--dir', dir_path, name))
            mocked_parse_json_arg.assert_not_called()
        mocked_package_class.browse.assert_called_once_with(name, None)
        mocked_package_class.assert_called_once_with()
        mocked_package = mocked_package_class.return_value
        mocked_package.set_dir.assert_called_once_with('.', dir_path, meta=expected_meta)
        mocked_package.push.assert_called_once_with(
            name, dest=None, message=None, registry=None, workflow=..., force=False, dedupe=False
        )


@pytest.mark.parametrize(
    'meta_data',
    [
        '{invalid: meta}',
        "{'single': 'quotation'}",
        '{"test": "meta", }',
    ]
)
def test_push_with_meta_data_error(
    meta_data,
    capsys
):
    name = 'test/name'

    with patch_package_class as mocked_package_class, \
         mock.patch('quilt3.main.parse_arg_json', wraps=main.parse_arg_json) as mocked_parse_json_arg:

        mocked_package_class.browse.side_effect = FileNotFoundError()

        with pytest.raises(SystemExit):
            main.main(('push', '--dir', '.', name, '--meta', meta_data))
        # check for expected stderr exception message
        captured = capsys.readouterr()
        assert 'is not a valid json string' in captured.err
        mocked_parse_json_arg.assert_called_once_with(meta_data)
        mocked_package_class.browse.assert_not_called()
        mocked_package_class.assert_not_called()


@pytest.mark.parametrize(
    'workflow_input, expected_workflow',
    [
        (None, ...),
        ('', None),
        ('test-workflow', 'test-workflow'),
    ]
)
def test_push_workflow(workflow_input, expected_workflow):
    name = 'test/name'
    dir_path = 'test/dir/path'

    with patch_package_class as mocked_package_class:
        mocked_package_class.browse.side_effect = FileNotFoundError()
        workflow_args = () if workflow_input is None else ('--workflow', workflow_input)
        main.main(('push', '--dir', dir_path, *workflow_args, name))

        mocked_package_class.assert_called_once_with()
        mocked_package = mocked_package_class.return_value
        mocked_package_class.browse.assert_called_once_with(name, None)
        mocked_package.set_dir.assert_called_once_with('.', dir_path, meta=None)
        mocked_package.push.assert_called_once_with(
            name, dest=None, message=None, registry=None, workflow=expected_workflow, force=False, dedupe=False
        )


def test_list_packages(capsys):
    registry = 's3://my_test_bucket/'
    pkg_names = ['foo/bar', 'foo/bar1', 'foo1/bar']
    with patch('quilt3.backends.s3.S3PackageRegistryV1.list_packages') as list_packages_mock:
        list_packages_mock.return_value = pkg_names
        main.main(('list-packages', registry))

        list_packages_mock.assert_called_once_with()
        captured = capsys.readouterr()
        assert captured.out.split() == pkg_names


def test_push_no_copy():
    name = 'test/name'
    dir_path = 's3://test/dir/path'

    with patch_package_class as mocked_package_class:
        main.main(('push', '--dir', dir_path, '--no-copy', name))

        mocked_package_class.browse.assert_called_once_with(name, None)
        mocked_package_class.assert_not_called()
        mocked_package = mocked_package_class.browse.return_value
        mocked_package.set_dir.assert_called_once_with('.', dir_path, meta=None)
        mocked_package.push.assert_called_once_with(
            name, registry=None, dest=None, message=None, workflow=..., force=False, dedupe=False,
            selector_fn=main._selector_fn_no_copy,
        )


def test_push_no_copy_local_dir(capsys):
    name = 'test/name'
    dir_path = 'test/dir/path'

    assert main.main(('push', '--dir', dir_path, '--no-copy', name)) == 1
    captured = capsys.readouterr()
    assert "--no-copy flag can be specified only for remote data." in captured.err



import operator
import os
import shutil
from datetime import datetime

from quilt3.data_transfer import delete_url

from .base import PackageRegistryV1, PackageRegistryV2


def safe_listdir(path):
    try:
        return os.listdir(path)
    except FileNotFoundError:
        return ()


class LocalPackageRegistryV1(PackageRegistryV1):
    def list_packages(self):
        pointers_dir_path = self.pointers_global_dir.path
        for usr in safe_listdir(pointers_dir_path):
            for pkg in safe_listdir(os.path.join(pointers_dir_path, usr)):
                yield f'{usr}/{pkg}'

    def list_package_pointers(self, pkg_name: str):
        pointers_dir_path = self.pointers_dir(pkg_name).path
        for path in safe_listdir(pointers_dir_path):
            with open(os.path.join(pointers_dir_path, path), encoding='utf-8') as f:
                yield path, f.read()

    def delete_package(self, pkg_name: str):
        shutil.rmtree(self._pointers_usr_dir(pkg_name).path)

    def delete_package_version(self, pkg_name: str, top_hash: str):
        super().delete_package_version(pkg_name, top_hash)
        delete_url(self.pointers_dir(pkg_name))
        delete_url(self._pointers_usr_dir(pkg_name))


class LocalPackageRegistryV2(PackageRegistryV2):
    _from_path_to_package_name = operator.methodcaller('replace', '@', '/')

    def list_packages(self):
        return map(self._from_path_to_package_name, safe_listdir(self.manifests_global_dir.path))

    list_package_pointers = LocalPackageRegistryV1.list_package_pointers

    def list_package_versions_with_timestamps(self, pkg_name: str):
        try:
            scanner = os.scandir(self.manifests_package_dir(pkg_name).path)
        except FileNotFoundError:
            return
        with scanner as entries:
            for entry in entries:
                yield datetime.fromtimestamp(entry.stat().st_mtime), entry.name

    def delete_package(self, pkg_name: str):
        shutil.rmtree(self.manifests_package_dir(pkg_name).path)
        shutil.rmtree(self.pointers_dir(pkg_name).path)

    def delete_package_version(self, pkg_name: str, top_hash: str):
        shutil.rmtree(self._manifest_parent_pk(pkg_name, top_hash).path)
        super().delete_package_version(pkg_name, top_hash)
        delete_url(self.pointers_dir(pkg_name))
        delete_url(self.manifests_package_dir(pkg_name))


def get_package_registry(version: int):
    if version == 1:
        return LocalPackageRegistryV1
    if version == 2:
        return LocalPackageRegistryV2
    raise ValueError()



import functools
import json
import re
import typing
from importlib import resources

import botocore.exceptions
import jsonschema
import yaml

from quilt3.data_transfer import get_bytes_and_effective_pk

from .. import util
from ..backends import PackageRegistry


class ConfigDataVersion(typing.NamedTuple):
    major: int
    minor: int
    patch: int

    @classmethod
    def parse(cls, version_str: str):
        """
        Parse valid version string.
        """
        return cls._make(((*map(int, version_str.split('.')), 0, 0)[:3]))

    def __str__(self):
        return '%s.%s.%s' % self


JSONSchemaError = typing.Union[jsonschema.ValidationError, jsonschema.SchemaError]


class WorkflowErrorBase(util.QuiltException):
    schema_validation_error: JSONSchemaError = None

    @classmethod
    def from_schema_validation_error(cls, message: str, err: JSONSchemaError):
        obj = cls(f'{message}: {err.message}.')
        obj.schema_validation_error = err
        return obj


class ConfigurationError(WorkflowErrorBase):
    pass


class UnsupportedConfigurationVersionError(ConfigurationError):
    def __init__(self, version: ConfigDataVersion):
        self.version = version
        super().__init__(f"Version '{version}' is not supported")


class WorkflowValidationError(WorkflowErrorBase):
    pass


@functools.lru_cache(maxsize=None)
def _get_conf_validator():
    schema = json.loads(resources.read_text(__name__, 'config-1.schema.json'))
    return jsonschema.Draft7Validator(schema).validate


SUPPORTED_META_SCHEMAS = {
    'http://json-schema.org/draft-07/schema#': jsonschema.Draft7Validator,
}


def _schema_load_object_hook(o):
    if '$ref' in o:
        raise util.QuiltException("Currently we don't support $ref in schema.")
    return o


_load_schema_json = json.JSONDecoder(object_hook=_schema_load_object_hook).decode


class WorkflowConfig:
    CONFIG_DATA_VERSION = ConfigDataVersion(1, 1, 0)

    def __init__(self, config: dict, physical_key: util.PhysicalKey):
        """
        Args:
            config: validated workflow config or `None` if there is no config
            physical_key: from where config was loaded
        """
        self.config = config
        self.physical_key = physical_key
        self.loaded_schemas_by_id = {}
        self.loaded_schemas = {}

    @staticmethod
    def get_config_data_version_str(data: dict) -> str:
        version_obj = data['version']
        assert isinstance(version_obj, (str, dict))
        return version_obj if isinstance(version_obj, str) else version_obj['base']

    @classmethod
    def is_supported_config_data_version(cls, version: ConfigDataVersion):
        return cls.CONFIG_DATA_VERSION >= version

    @classmethod
    def load(cls, pk: util.PhysicalKey):
        data = None
        try:
            data, pk = get_bytes_and_effective_pk(pk)
        except FileNotFoundError:
            pass
        except botocore.exceptions.ClientError as e:
            if e.response['Error']['Code'] != 'NoSuchKey':
                raise ConfigurationError(f"Couldn't load workflows config. {e}.")
        if data is None:
            return

        try:
            # TODO: raise if objects contain duplicate properties
            config = yaml.safe_load(data.decode())
        except yaml.YAMLError as e:
            raise ConfigurationError("Couldn't parse workflows config as YAML.") from e
        conf_validator = _get_conf_validator()
        try:
            conf_validator(config)
        except jsonschema.ValidationError as e:
            raise ConfigurationError.from_schema_validation_error('Workflows config failed validation', e) from e

        version_str = cls.get_config_data_version_str(config)
        version = ConfigDataVersion.parse(version_str)
        if not cls.is_supported_config_data_version(version):
            raise UnsupportedConfigurationVersionError(version)

        return cls(config, pk)

    def get_pk_for_schema_id(self, schema_id: str) -> util.PhysicalKey:
        schemas = self.config.get('schemas', {})
        if schema_id not in schemas:
            raise ConfigurationError(f'There is no {schema_id!r} in schemas.')
        schema_url = schemas[schema_id]['url']
        try:
            schema_pk = util.PhysicalKey.from_url(schema_url)
        except util.URLParseError as e:
            raise ConfigurationError(f"Couldn't parse URL {schema_url!r}.") from e
        if schema_pk.is_local() and not self.physical_key.is_local():
            raise ConfigurationError(f"Local schema {str(schema_pk)!r} can't be used on the remote registry.")

        return schema_pk

    def load_schema(self, schema_pk: util.PhysicalKey) -> (bytes, util.PhysicalKey):
        handled_exception = (OSError if schema_pk.is_local() else botocore.exceptions.ClientError)
        try:
            return get_bytes_and_effective_pk(schema_pk)
        except handled_exception as e:
            raise ConfigurationError(f"Couldn't load schema at {schema_pk}.") from e

    def make_validator_from_schema(self, schema_id):
        if schema_id in self.loaded_schemas_by_id:
            return self.loaded_schemas_by_id[schema_id][0]

        schema_pk = self.get_pk_for_schema_id(schema_id)
        if str(schema_pk) in self.loaded_schemas:
            self.loaded_schemas_by_id[schema_id] = self.loaded_schemas[str(schema_pk)]
            return self.loaded_schemas_by_id[schema_id][0]

        schema_data, schema_pk_to_store = self.load_schema(schema_pk)
        try:
            schema = _load_schema_json(schema_data.decode())
        except json.JSONDecodeError as e:
            raise ConfigurationError(f"Couldn't parse {schema_pk} as JSON.") from e

        validator_cls = jsonschema.Draft7Validator
        if isinstance(schema, dict) and '$schema' in schema:
            meta_schema = schema['$schema']
            if not isinstance(meta_schema, str):
                raise ConfigurationError('$schema must be a string.')
            validator_cls = SUPPORTED_META_SCHEMAS.get(meta_schema)
            if validator_cls is None:
                raise ConfigurationError(f"Unsupported meta-schema: {meta_schema}.")

        try:
            validator_cls.check_schema(schema)
        except jsonschema.SchemaError as e:
            raise ConfigurationError.from_schema_validation_error(f'Schema {schema_id!r} is not valid', e) from e

        validator = validator_cls(schema)
        self.loaded_schemas_by_id[schema_id] = self.loaded_schemas[str(schema_pk)] = (validator, schema_pk_to_store)
        return validator

    def get_workflow_validator(self, workflow):
        if workflow is ...:
            workflow = self.config.get('default_workflow')

        workflows_data = self.config['workflows']
        if workflow is None:
            if self.config.get('is_workflow_required', True):
                raise util.QuiltException('Workflow required, but none specified.')
            workflow_data = {}
        elif workflow not in workflows_data:
            raise util.QuiltException(f'There is no {workflow!r} workflow in config.')
        else:
            workflow_data = workflows_data[workflow]

        pkg_name_pattern = workflow_data.get('handle_pattern')
        pkg_name_pattern = re.compile(pkg_name_pattern) if pkg_name_pattern else None

        metadata_schema_id = workflow_data.get('metadata_schema')
        metadata_validator = self.make_validator_from_schema(metadata_schema_id) if metadata_schema_id else None

        entries_schema_id = workflow_data.get('entries_schema')
        entries_validator = self.make_validator_from_schema(entries_schema_id) if entries_schema_id else None

        is_message_required = workflow_data.get('is_message_required', False)

        data_to_store = {
            'id': workflow,
            'config': str(self.physical_key),
        }
        if self.loaded_schemas:
            data_to_store['schemas'] = {
                schema_id: str(x[1])
                for schema_id, x in self.loaded_schemas_by_id.items()
            }

        return WorkflowValidator(
            data_to_store=data_to_store,
            is_message_required=is_message_required,
            pkg_name_pattern=pkg_name_pattern,
            metadata_validator=metadata_validator,
            entries_validator=entries_validator,
        )


class WorkflowValidator(typing.NamedTuple):
    data_to_store: dict
    is_message_required: bool
    pkg_name_pattern: typing.Optional[typing.Pattern[str]]
    metadata_validator: typing.Any
    entries_validator: typing.Any

    def validate_name(self, name):
        if self.pkg_name_pattern and not self.pkg_name_pattern.search(name):
            raise WorkflowValidationError("Package name doesn't match required pattern.")

    def validate_message(self, message):
        if self.is_message_required and not message:
            raise WorkflowValidationError('Commit message is required by workflow, but none was provided.')

    def validate_metadata(self, meta):
        if self.metadata_validator is None:
            return
        try:
            self.metadata_validator.validate(meta)
        except jsonschema.ValidationError as e:
            raise WorkflowValidationError.from_schema_validation_error('Metadata failed validation', e) from e

    def validate_entries(self, pkg):
        if self.entries_validator is None:
            return
        try:
            self.entries_validator.validate(self.get_pkg_entries_for_validation(pkg))
        except jsonschema.ValidationError as e:
            raise WorkflowValidationError.from_schema_validation_error("Package entries failed validation", e) from e

    def get_pkg_entries_for_validation(self, pkg):
        # TODO: this should be validated without fully populating array.
        empty_dict = {}

        def reuse_empty_dict(meta):
            # Reuse the same empty dict for entries without meta
            # to reduce memory usage.
            return empty_dict if meta == {} else meta

        return [
            {
                'logical_key': lk,
                'size': e.size,
                "meta": reuse_empty_dict(e.meta),
            }
            for lk, e in pkg.walk()
        ]

    def validate(self, *, name, pkg, message):
        self.validate_message(message)
        self.validate_name(name)
        self.validate_metadata(pkg.meta)
        self.validate_entries(pkg)

        return self.data_to_store


def validate(*, registry: PackageRegistry, workflow, name, pkg, message):
    # workflow is ... => no workflow provided by user;
    # workflow is None => don't use any workflow.
    if not (workflow in (None, ...) or isinstance(workflow, str)):
        raise TypeError

    workflow_config = registry.get_workflow_config()
    if workflow_config is None:
        if workflow is ...:
            return
        raise util.QuiltException(f'{workflow!r} workflow is specified, but no workflows config exist.')

    workflow_validator = workflow_config.get_workflow_validator(workflow)
    return workflow_validator.validate(name=name, pkg=pkg, message=message)



import os
import sys
from pathlib import Path

from setuptools import find_packages, setup
from setuptools.command.install import install

VERSION = Path(Path(__file__).parent, "quilt3", "VERSION").read_text().strip()


def readme():
    return """\
Quilt manages data like code (with packages, repositories, browsing and
revision history) so that teams can experiment faster in machine learning,
biotech, and other data-driven domains.

The `quilt3` PyPI package allows you to build, push, and install data packages.
Visit the `documentation quickstart <https://docs.quiltdata.com/quickstart>`_
to learn more.
"""


class VerifyVersionCommand(install):
    """Custom command to verify that the git tag matches our version"""
    description = 'verify that the git tag matches our version'

    def run(self):
        tag = os.getenv('CIRCLE_TAG')

        if tag != VERSION:
            info = "Git tag: {0} does not match the version of this app: {1}".format(
                tag, VERSION
            )
            sys.exit(info)


setup(
    name="quilt3",
    version=VERSION,
    packages=find_packages(exclude=("tests", "tests.*")),
    description='Quilt: where data comes together',
    long_description=readme(),
    python_requires='>=3.7',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
    ],
    author='quiltdata',
    author_email='contact@quiltdata.io',
    license='Apache-2.0',
    url='https://github.com/quiltdata/quilt',
    keywords='',
    install_requires=[
        'platformdirs>=2',
        'aws-requests-auth>=0.4.2',
        'boto3>=1.10.0',
        'jsonlines==1.2.0',
        'PyYAML>=5.1',
        'requests>=2.12.4',
        'tenacity>=5.1.1',
        'tqdm>=4.32',
        'requests_futures==1.0.0',
        'jsonschema>=3,<5',
        'importlib_metadata; python_version < "3.8"',
    ],
    extras_require={
        'pyarrow': [
            'numpy>=1.14.0',                # required by pandas, but missing from its dependencies.
            'pandas>=0.19.2',
            'pyarrow>=0.14.1',              # as of 7/5/19: linux/circleci bugs on 0.14.0
        ],
        'anndata': ['anndata>=0.8.0'],
        'tests': [
            'anndata!=0.9.0',
            'quilt3[pyarrow,anndata]',
            'pytest==6.*',
            'pytest-cov',
            'coverage==6.4',
            'pytest-env',
            'pytest-subtests',
            'responses',
            'git-pylint-commit-hook',
        ],
        'catalog': [
            'quilt3_local>=1,<2',
            'uvicorn>=0.15,<0.18',
            # Workaround to reduce backtracking during `pip install quilt3[catalog]`.
            'aiobotocore[boto3]>=2',
        ],
    },
    include_package_data=True,
    entry_points={
        'console_scripts': ['quilt3=quilt3.main:main'],
    },
    cmdclass={
        'verify': VerifyVersionCommand,
    }
)
